'''Pure in-memory backend for kvlayer.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

The :class:`LocalStorage` backend is mostly useful in test environments:
it stores all :mod:`kvlayer` data in a Python dictionary.

.. autoclass:: AbstractLocalStorage
   :show-inheritance:

.. autoclass:: LocalStorage
   :show-inheritance:

'''
from __future__ import absolute_import
import logging
import time

from kvlayer._abstract_storage import AbstractStorage
from kvlayer._utils import _requires_connection

logger = logging.getLogger(__name__)


class AbstractLocalStorage(AbstractStorage):
    """Local in-memory storage for testing.

    This is a base class for storage implementations that use a
    dictionary-like object for actually holding data.

    """

    def __init__(self, *args, **kwargs):
        super(AbstractLocalStorage, self).__init__(*args, **kwargs)
        self._connected = False
        self._raise_on_missing = self._config.get('raise_on_missing', True)

    def setup_namespace(self, table_names, value_types={}):
        super(AbstractLocalStorage, self).setup_namespace(
            table_names, value_types)

        if self._app_name not in self._data:
            self._data[self._app_name] = {}
        if self._namespace not in self._data[self._app_name]:
            self._data[self._app_name][self._namespace] = {}

        for table in table_names:
            if table not in self.data:
                self.data[table] = dict()

        self._connected = True

    def delete_namespace(self):
        if ((self._app_name in self._data and
             self._namespace in self._data[self._app_name])):
            del self._data[self._app_name][self._namespace]
            if not self._data[self._app_name]:
                # empty now? del that too.
                del self._data[self._app_name]

    @property
    def data(self):
        '''The underlying dictionary for this namespace.'''
        return self._data[self._app_name][self._namespace]

    @_requires_connection
    def clear_table(self, table_name):
        self.data[table_name] = dict()

    @_requires_connection
    def put(self, table_name, *keys_and_values, **kwargs):
        start_time = time.time()
        keys_size = 0
        values_size = 0
        num_keys = 0
        for key, val in keys_and_values:
            key_spec = self._table_names[table_name]
            self.check_put_key_value(key, val, table_name, key_spec)
            self.data[table_name][key] = val
            if self._log_stats is not None:
                num_keys += 1
                keys_size += len(self._encoder.serialize(key, key_spec))
                values_size += len(str(val))

        end_time = time.time()
        num_values = num_keys

        self.log_put(table_name, start_time, end_time, num_keys, keys_size,
                     num_values, values_size)

    @_requires_connection
    def scan(self, table_name, *key_ranges, **kwargs):
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        num_values = 0
        values_size = 0

        key_spec = self._table_names[table_name]
        key_ranges = list(key_ranges)
        if not key_ranges:
            key_ranges = [[None, None]]
        for start, finish in key_ranges:
            total_count = 0
            start = self._encoder.make_start_key(start, key_spec)
            finish = self._encoder.make_end_key(finish, key_spec)
            for key in sorted(self.data[table_name].iterkeys()):
                # given a range, mimic the behavior of DBs that tell
                # you if they failed to find a key
                # LocalStorage does get/put on the Python tuple as the key,
                # stringify for sort comparison
                joined_key = self._encoder.serialize(key, key_spec)
                if (start is not None) and (start > joined_key):
                    continue
                if (finish is not None) and (finish < joined_key):
                    continue
                total_count += 1
                val = self.data[table_name][key]
                yield key, val

                if self._log_stats is not None:
                    num_keys += 1
                    keys_size += len(joined_key)
                    num_values += 1
                    values_size += len(str(val))

        end_time = time.time()
        num_values = num_keys
        self.log_scan(table_name, start_time, end_time, num_keys, keys_size,
                      num_values, values_size)

    @_requires_connection
    def get(self, table_name, *keys, **kwargs):
        start_time = time.time()
        num_keys = 0
        keys_size = 0
        num_values = 0
        values_size = 0

        for key in keys:
            if self._log_stats is not None:
                key_spec = self._table_names[table_name]
                num_keys += 1
                keys_size += len(self._encoder.serialize(key, key_spec))
            try:
                key, value = key, self.data[table_name][key]
                yield key, value
                num_values += 1
                values_size += len(str(value))
            except KeyError:
                yield key, None

        end_time = time.time()
        self.log_get(table_name, start_time, end_time, num_keys, keys_size,
                     num_values, values_size)

    @_requires_connection
    def delete(self, table_name, *keys):
        start_time = time.time()
        num_keys = 0
        keys_size = 0

        for key in keys:
            if self._log_stats is not None:
                key_spec = self._table_names[table_name]
                num_keys += 1
                keys_size += len(self._encoder.serialize(key, key_spec))
            self.data[table_name].pop(key, None)

        end_time = time.time()
        self.log_delete(table_name, start_time, end_time, num_keys, keys_size)

    def close(self):
        # prevent reading until connected again
        self._connected = False


class LocalStorage(AbstractLocalStorage):
    """Local in-memory storage for testing.

    All instances of LocalStorage share the same underlying data.

    As a special case, this ignores almost all of the configuration
    metadata, and can be constructed with no parameters, not even
    `app_name` or `namespace` (defaults will be provided to make
    the base class happy).

    """

    _data = {}

    def __init__(self, config=None, app_name=None, namespace=None,
                 *args, **kwargs):
        if config is None:
            config = {}
        if app_name is None and 'app_name' not in config:
            config['app_name'] = 'kvlayer'
        if namespace is None and 'namespace' not in config:
            config['namespace'] = 'kvlayer'
        super(LocalStorage, self).__init__(config=config, app_name=app_name,
                                           namespace=namespace, *args,
                                           **kwargs)



'''Key tuple encoder using ASCII text with percent escaping.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

.. autoclass:: AsciiPercentEncoder

'''
from __future__ import absolute_import
import struct
import uuid

from kvlayer.encoders.base import Encoder
from kvlayer._exceptions import BadKey, SerializationError


class PackedEncoder(Encoder):
    '''Key tuple encoder using compact binary representation.

    This serializes :class:`~uuid.UUID`, :class:`int`, and :class:`long`
    to 8-byte ASCII hex strings.  :class:`str` escapes ``%`` to ``%25``
    and null bytes to ``%00``.  Then the tuple of byte strings from
    these encodings is joined with null bytes between them.

    This encoding produces unreadable binary small keys.  Sort order
    is preserved except for strings containing null and percent bytes,
    and negative integers.

    '''
    config_name = 'packed'

    DELIMITER = '\0'

    def serialize(self, key, spec):
        '''Convert a key tuple to a database-native byte string.

        `key` must be a tuple at most as long as `spec`.  `spec` is a
        tuple of types, and each of the items in `key` must be an
        instance of the corresponding type in `spec`.  If this is
        not the case, :exc:`exceptions.TypeError` will be raised.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if spec is None:
            spec = (None,) * len(key)
        if len(key) > len(spec):
            raise BadKey('key too long, wanted {} parts, got {}'
                            .format(len(spec), len(key)))
        return b'\0'.join(self._gen_ser_parts(key, spec))

    def _gen_ser_parts(self, key, spec):
        for i in xrange(len(key)):
            frag = key[i]
            typ = spec[i]
            yield self._serialize_one(frag, typ, i)

    def _serialize_one(self, frag, typ, i):
        '''Serialize one fragment of a key tuple.

        If `typ` is :const:`None`, then any type of `frag` is
        acceptable.

        :param frag: part of the key
        :param type typ: expected type of `frag`
        :param int i: position in key spec list
        :return: serialized byte string
        :raise exceptions.TypeError: if `frag` has the wrong type
        :raise kvlayer._exceptions.SerializationError: if `typ` isn't
          something this can serialize

        '''
        if typ is not None and not isinstance(frag, typ):
            raise BadKey('expected {} in key but got {} ({!r})'
                            .format(typ, type(frag), frag))

        if typ == str:
            # Delimiter is '\0'
            # 'prefix' should sort before 'prefixsuffix'
            # ('\0', '\xff') should sort before ('\0\0','\xff')
            # but '\0\0\0\xff' would sort before '\0\0\xff'
            # SO, we must encode away all '\0' inside strings so that
            # delimiter '\0' always sorts before content.
            frag = frag.replace('\x01', '\x01\x03').replace('\x00', '\x01\x02')
            if i == 0:
                return frag
            else:
                assert len(frag) < 0x0ffff
                return frag + struct.pack('>H', len(frag))
        elif typ == int:
            assert frag <= 0x7fffffff
            return struct.pack('>I', frag + 0x80000000)
        elif (typ == long) or (typ == (int,long)) or (typ == (long,int)):
            assert frag < 0x7fffffffffffffff
            return struct.pack('>Q', frag + 0x8000000000000000)
        elif typ == uuid.UUID:
            return frag.get_bytes()
        else:
            raise BadKey("don't know how to serialize type {}".format(typ))

    def make_start_key(self, key, spec):
        '''Convert a partial key to the start of a scan range.

        This returns a byte string such that all key tuples greater
        than or equal to `key` :meth:`serialize` to greater than or
        equal to the returned byte string.  If `key` is shorter than
        `spec` then it recognizes all longer key tuples with the same
        prefix.  This generally works like :meth:`serialize`.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        dbkey = self.serialize(key, spec)
        #if spec and len(key) > 0 and len(key) < len(spec):
        #    dbkey += self.DELIMITER
        return dbkey

    def make_end_key(self, key, spec):
        '''Convert a partial key to the end of a scan range.

        This returns a byte string such that all key tuples less than
        or equal to `key` :meth:`serialize` to less than or equal to
        the returned byte string.  If `key` is shorter than `spec`
        then it recognizes all longer key tuples with the same prefix.
        This generally works like :meth:`serialize`.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        dbkey = self.serialize(key, spec)
        if spec:
            if len(key) == 0:
                dbkey += '\xff'
            elif len(key) < len(spec):
                dbkey += '\x01'
                #dbkey += chr(ord(self.DELIMITER) + 1)
                #dbkey = dbkey[:-1] + chr(ord(dbkey[-1]) + 1)
            else:
                dbkey += '\x00'
        else:
            dbkey += '\xff'
        return dbkey

    def deserialize(self, dbkey, spec):
        '''Convert a database-native byte string to a key tuple.

        `dbkey` is expected to have the same format as a call to
        :meth:`serialize` for a key as long as `spec`.  The return
        value will always have the same length as `spec`.

        :param bytes dbkey: key byte string from database
        :param tuple spec: key type tuple
        :return: deserialized key tuple
        :raise exceptions.ValueError: if `dbkey` is wrong
        :raise kvlayer._exceptions.SerializationError: if `spec`
          contains types that cannot be deserialized

        '''
        end = len(dbkey)
        i = len(spec) - 1
        parts = [None] * len(spec)
        while i >= 0:
            si = spec[i]
            if si == int:
                parts[i] = struct.unpack('>I', dbkey[end-4:end])[0] - 0x80000000
                end -= 4
            elif (si == long) or (si == (int,long)) or (si == (long,int)):
                parts[i] = struct.unpack('>Q', dbkey[end-8:end])[0] - 0x8000000000000000
                end -= 8
            elif si == uuid.UUID:
                parts[i] = uuid.UUID(bytes=dbkey[end-16:end])
                end -= 16
            elif si == str:
                if i == 0:
                    ts = dbkey[:end]
                else:
                    strlen = struct.unpack('>H', dbkey[end-2:end])[0]
                    end -= 2
                    ts = dbkey[end - strlen:end]
                    end -= strlen
                # TODO: check that there are no lone '\x01' bytes which would be an invalid encoding
                parts[i] = ts.replace('\x01\x02', '\x00').replace('\x01\x03', '\x01')
            else:
                raise BadKey("don't know how to decode key part type {}".format(si))
            end -= 1 # skip delimiter byte
            i -= 1
        return tuple(parts)



'''Base class for tuple-to-string encoders.

.. This software is released under an MIT/X11 open source license.
   Copyright 2015 Diffeo, Inc.

.. autoclass:: Encoder

'''
from __future__ import absolute_import
import abc


class Encoder(object):
    '''Base class for tuple-to-string encoders.

    .. automethod:: serialize
    .. automethod:: deserialize

    '''
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def serialize(self, key, spec):
        '''Convert a key tuple to a database-native byte string.

        `key` must be a tuple at most as long as `spec`.  `spec` is a
        tuple of types, and each of the items in `key` must be an
        instance of the corresponding type in `spec`.  This invariant
        is not checked, but see
        :meth:`kvlayer._abstract_storage.AbstractStorage.check_put_key_value`.
        If an encoder makes strong assumptions about types matching up
        then it should raise :exc:`TypeError`, but this is not
        guaranteed.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        pass

    def make_start_key(self, key, spec):
        '''Convert a partial key to the start of a scan range.

        This returns a byte string such that all key tuples greater
        than or equal to `key` :meth:`serialize` to greater than or
        equal to the returned byte string.  If `key` is shorter than
        `spec` then it recognizes all longer key tuples with the same
        prefix.  This generally works like :meth:`serialize` but may
        return slightly different values in some cases.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        return self.serialize(key, spec)

    def make_end_key(self, key, spec):
        '''Convert a partial key to the end of a scan range.

        This returns a byte string such that all key tuples less than
        or equal to `key` :meth:`serialize` to less than or equal to
        the returned byte string.  If `key` is shorter than `spec`
        then it recognizes all longer key tuples with the same prefix.
        This generally works like :meth:`serialize` but may return
        slightly different values in some cases.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        return self.serialize(key, spec)

    @abc.abstractmethod
    def deserialize(self, dbkey, spec):
        '''Convert a database-native byte string to a key tuple.

        `dbkey` is expected to have the same format as a call to
        :meth:`serialize` for a key as long as `spec`.  The return
        value will always have the same length as `spec`.

        :param bytes dbkey: key byte string from database
        :param tuple spec: key type tuple
        :return: deserialized key tuple
        :raise exceptions.ValueError: if `dbkey` is wrong
        :raise kvlayer._exceptions.SerializationError: if `spec`
          contains types that cannot be deserialized

        '''
        pass



'''
Definition of AbstractStorage, which all storage
implementations inherit.

Your use of this software is governed by your license agreement.

Copyright 2012-2015 Diffeo, Inc.
'''

from __future__ import absolute_import
import abc
import atexit
import collections
import itertools
import operator
import struct
import time
import uuid

from kvlayer.encoders import get_encoder
from kvlayer._exceptions import BadKey, ConfigurationError, ProgrammerError


class COUNTER(object):
    '''Integer counter value type.

    You cannot meaningfully instantiate this class.  Instead, pass
    it as a value in the `value_types` dictionary parameter to
    :meth:`AbstractStorage.setup_namespace`.

    If a table has this value type, then its values are integers,
    in the same way as if the table had value type
    :class:`int`; but you can also use the table as a counter
    using :meth:`AbstractStorage.increment`.

    '''


class ACCUMULATOR(object):
    '''Floating-point counter value type.

    You cannot meaningfully instantiate this class.  Instead, pass
    it as a value in the `value_types` dictionary parameter to
    :meth:`AbstractStorage.setup_namespace`.

    If a table has this value type, then its values are floating-point
    numbers, in the same way as if the table had value type
    :class:`float`; but you can also use the table as a counter using
    :meth:`AbstractStorage.increment`.

    '''


class AbstractStorage(object):
    '''Base class for all low-level storage implementations.

    All of the table-like structures we use are setup like this::

        namespace = dict(
            table_name = dict((UUID, UUID, ...): val)
            ...
        )

    where the number of UUIDs in the key is a configurable parameter
    of each table, and the "val" is always binary and might be a
    trivial value, like 1.

    '''
    __metaclass__ = abc.ABCMeta

    def check_put_key_value(self, key, value, table_name, key_spec=None,
                            value_type=None):
        '''Check that a key/value pair are consistent with the schema.

        :param tuple key: key to put
        :param value: value to put (ignored)
        :param str table_name: kvlayer table name (for errors only)
        :param tuple key_spec: definition of the table key
        :param value_type: type of the table value
        :raise kvlayer._exceptions.BadKey: if `key` doesn't match `key_spec`

        '''
        if key_spec is None:
            key_spec = self._table_names[table_name]
        if value_type is None:
            value_type = self._value_types[table_name]
        if value_type is COUNTER:
            value_type = int
        if value_type is ACCUMULATOR:
            value_type = float
        if not isinstance(key, tuple):
            raise BadKey('key should be tuple, but got %s' % (type(key),))
        if len(key) != len(key_spec):
            raise BadKey('%r wants %r parts in key tuple, but got %r' %
                         (table_name,  len(key_spec), len(key)))
        for kp, ks in zip(key, key_spec):
            if not isinstance(kp, ks):
                raise BadKey('part of key wanted type %s but got %s: %r' %
                             (ks, type(kp), kp))
        if not isinstance(value, value_type):
            raise BadKey('value should be %s, but got %s' %
                         (value_type, type(value)))

    def value_to_str(self, value, value_type):
        if value is None:
            return None
        if value_type is str:
            return value
        if value_type is int or value_type is COUNTER:
            return struct.pack('>i', value)
        if value_type is float or value_type is ACCUMULATOR:
            return struct.pack('>f', value)
        raise ConfigurationError('unexpected value_type {!r}'
                                 .format(value_type))

    def str_to_value(self, value, value_type):
        if value is None:
            return None
        if value_type is str:
            return value
        if value_type is int or value_type is COUNTER:
            return struct.unpack('>i', value)[0]
        if value_type is float or value_type is ACCUMULATOR:
            return struct.unpack('>f', value)[0]
        raise ConfigurationError('unexpected value_type {!r}'
                                 .format(value_type))

    def __init__(self, config, app_name=None, namespace=None):
        '''Initialize a storage instance with config dict.

        `config` is the configuration for this object; it may come
        from :mod:`yakonfig` or elsewhere.  If `app_name` or `namespace`
        are given as arguments to this method, they override the
        corresponding parameters in the configuration.  These two
        parameters _must_ be present, if not from the named arguments
        then in the configuration, otherwise :class:`ConfigurationError`
        will be raised.

        This understands the following keys in `config`:

        `app_name`
          wrapper name for all namespaces
        `namespace`
          wrapper name for all managed tables
        `storage_addresses`
          list of locations of data storage; backend-specific
        `log_stats`
          if provided, name of a file to which to log statistics
        `log_stats_interval_ops`
          if provided, log stats after this many operations
        `log_stats_interval_seconds`
          if provided, log stats after this many seconds
        `encoder`
          name of a key-to-string encoder, if applicable

        :param dict config: local configuration dictionary
        :param str app_name: optional app name override
        :param str namespace: optional namespace override
        :param kvlayer.encoders.base.Encoder encoder: key serializer
        :raise kvlayer._exceptions.ConfigurationError: if no `app_name`
          or `namespace` could be found
        '''
        self._config = config
        self._table_names = {}
        self._value_types = {}
        self._namespace = namespace or self._config.get('namespace', None)
        if not self._namespace:
            raise ConfigurationError('kvlayer requires a namespace')
        self._app_name = app_name or self._config.get('app_name', None)
        if not self._app_name:
            raise ConfigurationError('kvlayer requires an app_name')
        self._encoder = get_encoder(self._config.get('encoder', None))
        self._require_uuid = self._config.get('keys_must_be_uuid', True)
        log_stats_cfg = self._config.get('log_stats', None)
        # StorageStats also consumes:
        #  log_stats_interval_ops
        #  log_stats_interval_seconds
        if log_stats_cfg:
            self._log_stats = StorageStats(log_stats_cfg, self._config)
        else:
            self._log_stats = None

    def setup_namespace(self, table_names, value_types={}):
        '''Create tables in the namespace.

        Can be run multiple times with different `table_names` in
        order to expand the set of tables in the namespace.  This
        generally needs to be called by every client, even if only
        reading data.

        Tables are specified by the form of their keys. A key must be
        a tuple of a set number and type of parts. Currently types
        :class:`uuid.UUID`, :class:`int`, :class:`long`, and
        :class:`str` are well supported, anything else is serialzed by
        :func:`str`. Historically, a kvlayer key had to be a tuple of some
        number of UUIDs.  `table_names` is a dictionary mapping a
        table name to a tuple of types.  The dictionary values may also
        be integers, in which case the tuple is that many UUIDs.

        `value_types` specifies the type of the values for a given
        table.  Tables default to having a value type of :class:`str`.
        :class:`int` and :class:`float` are also permitted.  Value
        types may also be :class:`COUNTER` or :class:`ACCUMULATOR`;
        see :meth:`increment` for details on these types.  You must pass
        the corresponding type as the value parameter to :meth:`put`,
        and that type will be returned as the value part of
        :meth:`get` and :meth:`scan`.

        :param dict table_names: Mapping from table name to key type tuple
        :param dict value_types: Mapping from table name to value type

        '''
        # Subclass implementations should call this superclass
        # implementation to actually populate self._table_names.
        for k, v in table_names.iteritems():
            if isinstance(v, (int, long)):
                if v >= 50:
                    # This is probably a bug
                    raise ConfigurationError(
                        'excessively long tuple size {!r} for table {!r}'
                        .format(v, k))
                v = (uuid.UUID,) * v
            self._table_names[k] = v
            value_type = value_types.get(k, str)
            if value_type not in (str, int, float, COUNTER, ACCUMULATOR):
                raise ConfigurationError(
                    'invalid value type {!r} for table {!r}'
                    .format(value_type, k))
            self._value_types[k] = value_type

    @abc.abstractmethod
    def delete_namespace(self):
        '''Deletes all data from namespace.'''
        return

    @abc.abstractmethod
    def clear_table(self, table_name):
        '''Delete all data from one table.'''
        return

    @abc.abstractmethod
    def put(self, table_name, *keys_and_values, **kwargs):
        '''Save values for keys in `table_name`.

        Each key must be a tuple of length and types as specified for
        `table_name` in :meth:`setup_namespace`.

        '''
        return

    def log_put(self, table_name, start_time, end_time, num_keys, keys_size,
                num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.put.add(table_name, start_time, end_time, num_keys,
                                    keys_size, num_values, values_size)

    @abc.abstractmethod
    def scan(self, table_name, *key_ranges, **kwargs):
        '''Yield tuples of (key, value) from querying table_name for
        items with keys within the specified ranges.  If no key_ranges
        are provided, then yield all (key, value) pairs in table.
        This may return nothing if the table is empty or there are
        no matching keys in any of the specified ranges.

        Each of the `key_ranges` is a pair of a start and end tuple to
        scan.  To specify the beginning or end, a -Inf or Inf value,
        use an empty tuple as the beginning or ending key of a range.

        '''
        return

    def log_scan(self, table_name, start_time, end_time, num_keys, keys_size,
                 num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.scan.add(
                table_name, start_time, end_time,
                num_keys, keys_size, num_values, values_size)

    def scan_keys(self, table_name, *key_ranges, **kwargs):
        '''Scan only the keys from a table.

        Yields key tuples from queying `table_name` for keys within
        the specified ranges.  If no `key_ranges` are provided, then
        yield all key tuples in the table.  This may yield nothing if
        the table is empty or there are no matching keys in any of the
        specified ranges.

        Each of the `key_ranges` is a pair of a start and end tuple to
        scan.  To specify the beginning or end, a -Inf or Inf value,
        use an empty tuple as the beginning or ending key of a range.

        '''
        # Feel free to reimplement this if your backend can do better!
        # we don't do log_scan_keys() here, let underlying scan() call
        # log_scan()
        return itertools.imap(operator.itemgetter(0),
                              self.scan(table_name, *key_ranges, **kwargs))

    def log_scan_keys(self, table_name, start_time, end_time, num_keys,
                      keys_size):
        if self._log_stats is not None:
            self._log_stats.scan_keys.add(table_name, start_time, end_time,
                                          num_keys, keys_size, 0, 0)

    @abc.abstractmethod
    def get(self, table_name, *keys, **kwargs):

        '''Yield tuples of (key, value) from querying table_name for items
        with keys.  If any of the key tuples are not in the table,
        those key tuples will be yielded with value :const:`None`.

        '''
        return

    def log_get(self, table_name, start_time, end_time, num_keys, keys_size,
                num_values, values_size):
        if self._log_stats is not None:
            self._log_stats.get.add(
                table_name, start_time, end_time,
                num_keys, keys_size, num_values, values_size)

    @abc.abstractmethod
    def delete(self, table_name, *keys, **kwargs):
        '''Delete all (key, value) pairs with specififed keys

        '''
        return

    def log_delete(self, table_name, start_time, end_time, num_keys,
                   keys_size):
        if self._log_stats is not None:
            self._log_stats.delete.add(table_name, start_time, end_time,
                                       num_keys, keys_size, 0, 0)

    @abc.abstractmethod
    def close(self):
        '''
        close connections and end use of this storage client
        '''
        if self._log_stats is not None:
            self._log_stats.close()
            self._log_stats = None

    def increment(self, table_name, *keys_and_values):
        '''Add values to a counter-type table.

        `keys_and_values` are parameters of `(key, value)` pairs.  The
        values must be :class:`int`, if `table_name` is a
        :class:`COUNTER` table, or :class:`float`, if `table_name` is
        a :class:`ACCUMULATOR` table.  For each key, the current value
        is fetched from the storage, the value is added to it, and the
        resulting value added back into the storage.

        This method is not guaranteed to be atomic, either on a
        specific key or across all keys, but specific backends may
        have better guarantees.  The behavior is unspecified if the
        same key is included multiple times in the parameter list.

        To use this, you must have passed `table_name` to
        :meth:`setup_namespaces` in its `value_types` parameter,
        setting the value type to :data:`COUNTER` or
        :data:`ACCUMULATOR`.  When you do this, you pass and receive
        :class:`float` values back from all methods in this class for
        that table.  :meth:`put` directly sets the values of counter
        keys.  Counter values default to 0; if you change a counter
        value to 0 then it will be "present" for purposes of
        :meth:`get`, :meth:`scan`, and :meth:`scan_keys`.

        :param str table_name: name of table to update
        :param keys_and_values: additional parameters are pairs of
          key tuple and numeric delta value

        '''
        if self._value_types[table_name] not in [COUNTER, ACCUMULATOR]:
            raise ProgrammerError('table {} is not a counter table'
                                  .format(table_name))
        # Default, non-atomic implementation
        keys = [k for (k, v) in keys_and_values]
        deltas = [v for (k, v) in keys_and_values]
        old_keys_values = self.get(table_name, *keys)
        new_keys_values = [(k, (v or 0)+d)
                           for ((k, v), d) in zip(old_keys_values, deltas)]
        self.put(table_name, *new_keys_values)


class StringKeyedStorage(AbstractStorage):
    '''Partial implementation of AbstractStorage using string keys.

    This assumes that the underlying database only can deal in byte-string
    keys, and :attr:`encoder` needs to be used to generate keys and values.
    This provides wrappers that do the encoding and basic stats gathering,
    requiring derived classes to only provide the underlying machinery.

    '''

    def put(self, table_name, *keys_and_values, **kwargs):
        start_time = time.time()
        for (k, v) in keys_and_values:
            self.check_put_key_value(k, v, table_name)
        ks = [self._encoder.serialize(k, self._table_names[table_name])
              for (k, v) in keys_and_values]
        vs = [self.value_to_str(v, self._value_types[table_name])
              for (k, v) in keys_and_values]
        self._put(table_name, zip(ks, vs), **kwargs)
        end_time = time.time()
        if self._log_stats is not None:
            self._log_stats.put.add(table_name, start_time, end_time,
                                    len(ks), sum(len(k) for k in ks),
                                    len(vs), sum(len(v) for v in vs))

    @abc.abstractmethod
    def _put(self, table_name, keys_and_values):
        pass

    def scan(self, table_name, *key_ranges, **kwargs):
        stats = StatRecord()
        key_spec = self._table_names[table_name]
        value_type = self._value_types[table_name]
        new_key_ranges = [(self._encoder.make_start_key(start, key_spec),
                           self._encoder.make_end_key(end, key_spec))
                          for (start, end) in key_ranges]
        for k, v in self._scan(table_name, new_key_ranges, **kwargs):
            stats.record(len(k), len(v))
            yield (self._encoder.deserialize(k, key_spec),
                   self.str_to_value(v, value_type))
        if self._log_stats is not None:
            self._log_stats.scan.add_rec(table_name, stats)

    @abc.abstractmethod
    def _scan(self, table_name, key_ranges):
        pass

    def scan_keys(self, table_name, *key_ranges, **kwargs):
        stats = StatRecord()
        key_spec = self._table_names[table_name]
        new_key_ranges = [(self._encoder.make_start_key(start, key_spec),
                           self._encoder.make_end_key(end, key_spec))
                          for (start, end) in key_ranges]
        for k in self._scan_keys(table_name, new_key_ranges, **kwargs):
            stats.record(len(k), None)
            yield self._encoder.deserialize(k, key_spec)
        if self._log_stats is not None:
            self._log_stats.scan_keys.add_rec(table_name, stats)

    def _scan_keys(self, table_name, key_ranges):
        for (k, v) in self._scan(table_name, key_ranges):
            yield k

    def get(self, table_name, *keys, **kwargs):
        stats = StatRecord()
        key_spec = self._table_names[table_name]
        value_type = self._value_types[table_name]
        new_keys = [self._encoder.serialize(k, key_spec) for k in keys]
        for (k, v) in self._get(table_name, new_keys, **kwargs):
            if v is None:
                stats.record(len(k), None)
            else:
                stats.record(len(k), len(v))
            yield (self._encoder.deserialize(k, key_spec),
                   self.str_to_value(v, value_type))
        if self._log_stats is not None:
            self._log_stats.get.add_rec(table_name, stats)

    @abc.abstractmethod
    def _get(self, table_name, keys):
        pass

    def delete(self, table_name, *keys, **kwargs):
        start_time = time.time()
        key_spec = self._table_names[table_name]
        new_keys = [self._encoder.serialize(k, key_spec) for k in keys]
        self._delete(table_name, new_keys, **kwargs)
        if self._log_stats is not None:
            self._log_stats.delete.add(
                table_name, start_time, time.time(),
                len(new_keys), sum(len(k) for k in new_keys), 0, 0)

    @abc.abstractmethod
    def _delete(self, table_name, keys):
        pass


# I may have built this inside-out.
# Maybe the top level split should be on table, and keep stats of ops within
# that.  It shouldn't be hard to transpose for display if needed.
class StorageStats(object):
    def __init__(self, config_str_or_writeable, config):
        self._f = None
        self._config_str = None
        if hasattr(config_str_or_writeable, 'write'):
            self._f = config_str_or_writeable
        elif isinstance(config_str_or_writeable, (str, unicode)):
            self._config_str = config_str_or_writeable
        self._interval_ops = config.get('log_stats_interval_ops')
        if self._interval_ops is not None:
            self._interval_ops = int(self._interval_ops)
            if self._interval_ops <= 0:
                self._interval_ops = None
        self._op_interval_counter = 0
        self._interval_seconds = config.get('log_stats_interval_seconds')
        if self._interval_seconds is not None:
            self._interval_seconds = float(self._interval_seconds)
            if self._interval_seconds <= 0.0:
                self._interval_seconds = None
        self._interval_s_last_flushed = time.time()
        self.put = OpStats(self)
        self.scan = OpStats(self)
        self.scan_keys = OpStats(self)
        self.get = OpStats(self)
        self.delete = OpStats(self)

        self._closed = False
        atexit.register(self.atexit)

    def __str__(self):
        return ('put:\n{x.put}\nscan:\n{x.scan}\nscan_keys:\n{x.scan_keys}\n'
                'get:\n{x.get}\ndelete:\n{x.delete}\n'.format(x=self))

    def _out(self):
        if (self._f is None) and hasattr(self._config_str, 'write'):
            self._f = self._config_str
        if self._f is None:
            self._f = open(self._config_str, 'a')
        return self._f

    def did_op(self):
        if self._closed:
            return
        if self._interval_ops is not None:
            self._op_interval_counter += 1
            if self._op_interval_counter >= self._interval_ops:
                self.flush()
                self._interval_s_last_flushed = time.time()
                self._op_interval_counter = 0
                return
        if self._interval_seconds is not None:
            now = time.time()
            if self._interval_s_last_flushed + self._interval_seconds < now:
                self.flush()
                self._interval_s_last_flushed = now
                self._op_interval_counter = 0
                return

    def flush(self):
        out = self._out()
        out.write(time.strftime('%Y%m%d_%H%M%S\n'))
        out.write(self.__str__())
        if hasattr(out, 'flush'):
            out.flush()

    def close(self):
        if self._f is not None:
            if hasattr(self._f, 'close'):
                self._f.close()
            self._f = None
        self._closed = True

    def atexit(self):
        if not self._closed:
            self.flush()
            self.close()


class OpStats(object):
    def __init__(self, parentStorageStats):
        self.by_table = collections.defaultdict(OpStatsPerTable)
        self.num_ops = 0
        self._psto = parentStorageStats

    def add(self, table_name, start_time, end_time, num_keys, keys_size,
            num_values, values_size):
        self.num_ops += 1
        self.by_table[table_name].add(start_time, end_time,
                                      num_keys, keys_size,
                                      num_values, values_size)
        self._psto.did_op()

    def add_rec(self, table_name, record):
        '''Variant of :meth:`add` taking a :class:`StatRecord`.'''
        return self.add(table_name, record.start_time, time.time(),
                        record.num_keys, record.keys_size,
                        record.num_values, record.values_size)

    def __str__(self):
        parts = []
        total = OpStatsPerTable()
        for k, v in self.by_table.iteritems():
            total += v
            parts.append('{:10s} {}\n'.format(k, str(v)))
        return ''.join(parts) + '           {}\n'.format(str(total))


class OpStatsPerTable(object):
    def __init__(self):
        self.total_time = 0.0
        self.num_ops = 0
        self.num_keys = 0
        self.keys_size = 0
        self.num_values = 0
        self.values_size = 0

    def add(self, start_time, end_time, num_keys, keys_size, num_values,
            values_size=0):
        self.total_time += (end_time - start_time)
        self.num_ops += 1
        self.num_keys += num_keys
        self.keys_size += keys_size
        self.num_values += num_values
        self.values_size += values_size

    def __iadd__(self, b):
        assert isinstance(b, OpStatsPerTable)
        self.total_time += b.total_time
        self.num_ops += b.num_ops
        self.num_keys += b.num_keys
        self.keys_size += b.keys_size
        self.num_values += b.num_values
        self.values_size += b.values_size
        return self

    def __str__(self):
        out = ('{t:0.3f}s on {ops} ops ({spops:0.6g} s/op), '
               '{k} keys for {kb} bytes ({bpk} B/key)'.format(
                   t=self.total_time,
                   ops=self.num_ops,
                   spops=self.num_ops and (self.total_time/self.num_ops),
                   k=self.num_keys,
                   kb=self.keys_size,
                   bpk=self.num_keys and ((1.0*self.keys_size)/self.num_keys)
               ))
        if self.num_values > 0:
            out += '{v} values for {vb} bytes ({bpv:0.1f} B/val)'.format(
                v=self.num_values,
                vb=self.values_size,
                bpv=(1.0*self.values_size)/self.num_values
            )
        if self.total_time > 0:
            out += ' {:0.1f} (k+v)B/s'.format(
                (1.0*(self.keys_size + self.values_size))/self.total_time)
        return out


class StatRecord(object):
    '''Combined statistics record for a single action.'''

    def __init__(self):
        #: Starting time of the action
        self.start_time = time.time()
        #: Number of keys processed
        self.num_keys = 0
        #: Total byte size of keys processed
        self.keys_size = 0
        #: Number of values processed
        self.num_values = 0
        #: Total byte size of values processed
        self.values_size = 0

    def record(self, key_size, value_size):
        '''Add a single item to the record.

        Either `key_size` or `value_size` may be :const:`None`; if so then
        don't count that part.  :meth:`AbstractStorage.scan_keys`
        implementations will typically pass :const:`None` for `value_size`,
        for instance.  Otherwise increment the relevant counts and add
        the relevant byte sizes.

        '''
        if key_size is not None:
            self.num_keys += 1
            self.keys_size += key_size
        if value_size is not None:
            self.num_values += 1
            self.values_size += value_size



#!/usr/bin/env python
"""kvlayer self-tests.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

"""
from __future__ import absolute_import
import argparse
import os
import subprocess
import sys

try:
    import pytest
except ImportError:
    pytest = None

import kvlayer.tests.performance as performance_tests


def main():
    parser = argparse.ArgumentParser(description='Run all kvlayer tests.')
    if pytest:
        parser.add_argument('--unit', action='store_true',
                            help='Run unit tests')
    parser.add_argument('--performance', '--perf', action='store_true',
                        help='Run performance tests')
    parser.add_argument('redis_address', metavar='HOST:PORT',
                        help='location of a redis instance to use for testing')
    args = parser.parse_args()

    if not args.unit and not args.performance:
        args.unit = (pytest is not None)
        args.performance = True

    rc = 0

    if args.unit:
        # Since the performance tests do complicated multiprocessing
        # things, and things like the pytest-cov extension register
        # multiprocessing hooks to cooperate with pytest-xdist, make
        # sure py.test runs in an isolated process environment.
        ret = subprocess.call([sys.executable, '-m', 'py.test',
                               '--runslow', '--runperf',
                               '--redis-address', args.redis_address,
                               '-k', 'not (cassandra or accumulo or postgres or cborproxy or hbase or riak or mysql or postgrest)',
                               '-v',
                               os.path.dirname(__file__)])
        if rc == 0:
            rc = ret

    if args.performance:
        ret = performance_tests.run_all_perftests(args.redis_address, clientlist=['redis', 'local'])
        if rc == 0:
            rc = ret

    return rc

if __name__ == '__main__':
    sys.exit(main())



'''Implementation of AbstractStorage using a proxy server running a CBOR RPC protocol

CBOR RPC server for talking to Java databases like Accumulo:
https://github.com/diffeo/kvlayer-java-proxy

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''

from __future__ import absolute_import
import logging
import random
import socket
import time

import cbor

from kvlayer._abstract_storage import StringKeyedStorage
from kvlayer._exceptions import ProgrammerError

logger = logging.getLogger(__name__)


class SocketReader(object):
    '''
    Simple adapter from socket.recv to file-like-read
    '''
    def __init__(self, sock):
        self.socket = sock
        self.timeout_seconds = 10.0

    def read(self, num):
        start = time.time()
        data = self.socket.recv(num)
        while len(data) < num:
            now = time.time()
            if now > (start + self.timeout_seconds):
                break
            ndat = self.socket.recv(num - len(data))
            if ndat:
                data += ndat
        return data


class CborRpcClient(object):
    '''Base class for all client objects.

    This provides common `addr_family`, `address`, and `registry_addresses`
    configuration parameters, and manages the connection back to the server.

    Automatic retry and time based fallback is managed from
    configuration parameters `retries` (default 5), and
    `base_retry_seconds` (default 0.5). Retry time doubles on each
    retry. E.g. try 0; wait 0.5s; try 1; wait 1s; try 2; wait 2s; try
    3; wait 4s; try 4; wait 8s; try 5; FAIL. Total time waited just
    under base_retry_seconds * (2 ** retries).

    .. automethod:: __init__
    .. automethod:: _rpc
    .. automethod:: close

    '''

    def __init__(self, config=None):
        self._socket_family = config.get('addr_family', socket.AF_INET)
        # may need to be ('host', port)
        self._socket_addr = config.get('address')
        if self._socket_family == socket.AF_INET:
            if not isinstance(self._socket_addr, tuple):
                # python socket standard library insists this be tuple!
                tsocket_addr = tuple(self._socket_addr)
                assert len(tsocket_addr) == 2, ('address must be length-2 '
                                                'tuple ("hostname", '
                                                'port number), got {!r} '
                                                'tuplified to {!r}'
                                                .format(self._socket_addr,
                                                        tsocket_addr))
                self._socket_addr = tsocket_addr
        self._socket = None
        self._rfile = None
        self._local_addr = None
        self._message_count = 0
        self._retries = config.get('retries', 5)
        self._base_retry_seconds = float(config.get('base_retry_seconds', 0.5))

    def _conn(self):
        # lazy socket opener
        if self._socket is None:
            try:
                self._socket = socket.create_connection(self._socket_addr)
                self._local_addr = self._socket.getsockname()
            except:
                logger.error('error connecting to %r:%r', self._socket_addr[0],
                             self._socket_addr[1], exc_info=True)
                raise
        return self._socket

    def close(self):
        '''Close the connection to the server.

        The next RPC call will reopen the connection.

        '''
        if self._socket is not None:
            self._rfile = None
            try:
                self._socket.shutdown(socket.SHUT_RDWR)
                self._socket.close()
            except socket.error:
                logger.warn('error closing lockd client socket',
                            exc_info=True)
            self._socket = None

    @property
    def rfile(self):
        if self._rfile is None:
            conn = self._conn()
            self._rfile = SocketReader(conn)
        return self._rfile

    def _rpc(self, method_name, params):
        '''Call a method on the server.

        Calls ``method_name(*params)`` remotely, and returns the results
        of that function call.  Expected return types are primitives, lists,
        and dictionaries.

        :raise Exception: if the server response was a failure

        '''
        mlog = logging.getLogger('cborrpc')
        tryn = 0
        delay = self._base_retry_seconds
        self._message_count += 1
        message = {
            'id': self._message_count,
            'method': method_name,
            'params': params
        }
        mlog.debug('request %r', message)
        buf = cbor.dumps(message)

        errormessage = None
        while True:
            try:
                conn = self._conn()
                conn.send(buf)
                response = cbor.load(self.rfile)
                mlog.debug('response %r', response)
                assert response['id'] == message['id']
                if 'result' in response:
                    return response['result']
                # From here on out we got a response, the server
                # didn't have some weird intermittent error or
                # non-connectivity, it gave us an error message. We
                # don't retry that, we raise it to the user.
                errormessage = response.get('error')
                if errormessage and hasattr(errormessage, 'get'):
                    errormessage = errormessage.get('message')
                if not errormessage:
                    errormessage = repr(response)
                break
            except Exception as ex:
                if tryn < self._retries:
                    tryn += 1
                    logger.debug('ex in %r (%s), retrying %s in %s sec...',
                                 method_name, ex, tryn, delay, exc_info=True)
                    self.close()
                    time.sleep(delay)
                    delay *= 2
                    continue
                logger.error('failed in rpc %r %r', method_name, params,
                             exc_info=True)
                raise
        raise Exception(errormessage)


# def _rangesk(key, key_spec):
#     if not key:
#         return None
#     return serialize_key(key, key_spec=key_spec)


class CborProxyStorage(StringKeyedStorage):
    def __init__(self, *args, **kwargs):
        super(CborProxyStorage, self).__init__(*args, **kwargs)

        zk_addresses = self._config.get('zk_addresses', None)
        if not zk_addresses:
            raise ProgrammerError('config lacks zk_addresses')
        proxy_addresses = self._config.get('proxy_addresses', None)
        if not proxy_addresses:
            raise ProgrammerError('config lacks proxy_addresses')

        def str_to_pair(address):
            if ':' not in address:
                return (address, 7123)
            else:
                h, p = address.split(':')
                return (h, int(p))
        self._zk_addresses = zk_addresses  # map(str_to_pair, zk_addresses)
        self._proxy_addresses = map(str_to_pair, proxy_addresses)

        # cached lazy connection
        self._conn = None
        # have we sent zk connection information to proxy?
        self._proxy_connected = False

    @property
    def conn(self):
        if not self._conn:
            host, port = random.choice(self._proxy_addresses)
            logger.debug('connecting to cbor proxy %s:%s', host, port)
            self._conn = CborRpcClient({'address': (host, port)})
            zk_addr = random.choice(self._zk_addresses)
            ok, msg = self._conn._rpc(
                u'connect',
                [unicode(zk_addr),
                 unicode(self._config.get('username')),
                 unicode(self._config.get('password'))])
            if not ok:
                raise Exception(msg)
        return self._conn

    def _ns(self, table):
        return '%s_%s_%s' % (self._app_name, self._namespace, table)

    def setup_namespace(self, table_names, value_types):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.
        '''
        super(CborProxyStorage, self).setup_namespace(table_names, value_types)
        simple_table_names = {self._ns(k): dict()
                              for k in table_names.iterkeys()}
        self.conn._rpc(u'setup_namespace', [simple_table_names])

    def delete_namespace(self):
        simple_table_names = {self._ns(k): dict()
                              for k in self._table_names.iterkeys()}
        self.conn._rpc(u'delete_namespace', [simple_table_names])

    def clear_table(self, table_name):
        table_name = self._ns(table_name)
        self.conn._rpc(u'clear_table', [unicode(table_name)])

    def _put(self, table_name, keys_and_values):
        self.conn._rpc(u'put',
                       [unicode(self._ns(table_name)), keys_and_values])

    def _scan(self, table_name, key_ranges):
        table_name = self._ns(table_name)
        if not key_ranges:
            key_ranges = [((), ())]
        return self.conn._rpc(u'scan', [unicode(table_name), key_ranges])

    def _scan_keys(self, table_name, key_ranges):
        table_name = self._ns(table_name)
        if not key_ranges:
            key_ranges = [((), ())]
        return self.conn._rpc(u'scan_keys', [unicode(table_name), key_ranges])

    def _get(self, table_name, keys):
        table_name = self._ns(table_name)
        return self.conn._rpc(u'get', [unicode(table_name), keys])

    def close(self):
        if self._conn:
            self._conn.close()
            self._conn = None

    def _delete(self, table_name, keys, **kwargs):
        table_name = self._ns(table_name)
        self.conn._rpc(u'delete', [unicode(table_name), keys])

    def shutdown_proxies(self):
        for host, port in self._proxy_addresses:
            xc = CborRpcClient({'address': (host, port), 'retries': 0})
            try:
                xc._rpc(u"shutdown", [])
            except EOFError:
                pass  # ok
            except:
                logger.info("%s:%s shutdown", host, port, exc_info=True)



'''Exceptions for :mod:`kvlayer`.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''

class KVLayerError(Exception):
    '''Base class for all :mod:`kvlayer` exceptions.'''
    pass


class ConfigurationError(KVLayerError):
    '''An invalid value was passed to an object.

    The value is probably something user-provided, such as the name of
    the storage backend.  This error is raised if the invalid value is
    not detected until code is being executed.  If the value can be
    detected during the :mod:`yakonfig` configuration phase, then
    :class:`yakonfig.ConfigurationError` will be raised there instead.

    '''
    pass


class ProgrammerError(KVLayerError):
    '''The API was used incorrectly.'''
    pass


class StorageClosed(KVLayerError):
    '''A storage object was used after it was explicitly closed.'''
    pass


class DatabaseEmpty(KVLayerError):
    pass


class BadKey(KVLayerError):
    """A key value passed to a kvlayer function was not of the correct form.

    Keys must be tuples of a fixed length and with specific types.
    The length of the tuple is specified in the initial call to
    :meth:`kvlayer._abstract_storage.AbstractStorage.setup_namespace`.

    """
    pass


class SerializationError(KVLayerError):
    """Converting between an item and a serialized form failed."""
    pass



'''kvlayer backend for Riak.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014-2015 Diffeo, Inc.

'''
from __future__ import absolute_import

import riak

from kvlayer._abstract_storage import StringKeyedStorage


class RiakStorage(StringKeyedStorage):
    def __init__(self, *args, **kwargs):
        '''Create a new Riak client object.

        This sets up a :class:`riak.RiakStorage` according to configuration,
        but in general this will not result in a network connection.

        '''
        super(RiakStorage, self).__init__(*args, **kwargs)

        def make_node(s):
            if isinstance(s, basestring):
                return {'host': s}
            return s
        nodes = [make_node(s) for s in self._config['storage_addresses']]
        self.connection = riak.RiakClient(
            protocol=self._config.get('protocol', 'pbc'),
            nodes=nodes)
        self.scan_limit = self._config.get('scan_limit', 100)

    def _bucket(self, table):
        '''Riak bucket name for a kvlayer table.'''
        name = '{}_{}_{}'.format(self._app_name, self._namespace, table)
        return self.connection.bucket(name)

    def delete_namespace(self):
        '''Deletes all data from the namespace.

        This only actually deletes keys in known namespaces, as per
        :meth:`setup_namespace`.  It needs to iterate and individually
        delete every single key.

        '''
        for table in self._table_names.iterkeys():
            self.clear_table(table)

    def clear_table(self, table_name):
        '''Deletes all data from a single table.

        This needs to iterate and delete every single key in the
        corresponding Riak bucket.

        '''
        # See the riak-users thread around
        # http://permalink.gmane.org/gmane.comp.db.riak.user/14411
        # This strikes me as a little unsafe (shouldn't we push a vclock
        # into the delete operation?) but it's what that thread endorses
        bucket = self._bucket(table_name)
        for k in bucket.get_keys():
            bucket.delete(k)

    def _put(self, table_name, keys_and_values):
        '''Write some data to a table.

        Because of the way Riak works, each key/value pair is a separate
        write operation.  This backend makes no distinction between one
        write with multiple keys and multiple writes with one key.

        :param keys_and_values: data items to write
        :paramtype keys_and_values: pairs of (key, value)

        '''
        bucket = self._bucket(table_name)
        for k, v in keys_and_values:
            # Always do this with a read/write to maintain vector clock
            # consistency...even though this means we're pushing objects
            # around more than we need to
            obj = bucket.get(k)
            obj.encoded_data = v
            obj.content_type = 'application/octet-stream'
            obj.store()

    def _scan(self, table_name, key_ranges):
        '''Scan key/value ranges from a table.

        This is not a native Riak operation!  It is implemented as an
        index scan, over the special index ``$key``.

        '''
        return self._do_scan(table_name, key_ranges, with_values=True)

    def _scan_keys(self, table_name, key_ranges):
        '''Scan key ranges from a table.

        This is not a native Riak operation!  It is implemented as an
        index scan, over the special index ``$key``.

        '''
        return self._do_scan(table_name, key_ranges, with_values=False)

    def _do_scan(self, table_name, key_ranges, with_values=False):
        bucket = self._bucket(table_name)

        # Can this be a map/reduce job?  This would save us from the
        # requirement to keep a secondary index duplicating the key,
        # and correspondingly the requirement to use leveldb.  But,
        # the documentation notes in big letters, "Riak MapReduce is
        # intended for batch processing, not real time querying", and
        # there are some comments elsewhere in the documentation about
        # limitations of it (map jobs always query with effective R=1,
        # for instance).  In particular MR appears to always circulate
        # the key list, which makes it bad for large scans.

        if not key_ranges:
            key_ranges = [(None, None)]
        for start_key, end_key in key_ranges:
            if not start_key:
                start_key = b'\0'
            if not end_key:
                end_key = b'\xff'

            results = bucket.get_index('$key', startkey=start_key,
                                       endkey=end_key,
                                       max_results=self.scan_limit)
            while True:
                for key in results:
                    # Contrary to what the Riak documentation claims,
                    # in practice the $key and $bucket indexes seem
                    # to contain every key that ever existed.  That
                    # means we must do a fetch to ensure the key
                    # really exists.
                    obj = bucket.get(key)
                    if obj.exists:
                        if with_values:
                            yield (key, obj.encoded_data)
                        else:
                            yield key
                # NB: work around a bug in the Riak 1.4.9 protobuf
                # client.  If this is the last page of results, it
                # sets results.continuation='', but
                # IndexPage.has_next_page() tests "is None".  We
                # really want to be calling has_next_page().
                if results.continuation:  # results.has_next_page():
                    results = results.next_page()
                else:
                    break

    def _get(self, table_name, keys):
        '''Yield tuples of (key, value) for specific keys.'''
        bucket = self._bucket(table_name)

        # We can, in principle, use bucket.multiget() here.  That's
        # a complicated thing that fires up a thread pool under the
        # hood for lots of concurrent fetches.  In practice, get()
        # key lists are almost always pretty small.

        for key in keys:
            obj = bucket.get(key)
            if obj.exists:
                yield (key, obj.encoded_data)
            else:
                yield (key, None)

    def _delete(self, table_name, keys):
        '''Delete some specific keys.'''
        bucket = self._bucket(table_name)

        for key in keys:
            # Always do this with a read/write to maintain vector clock
            # consistency...even though this means we're pushing objects
            # around more than we need to
            obj = bucket.get(key)
            obj.delete()

    def close(self):
        '''End use of this storage client.

        While the Python Riak client maintains an internal connection
        pool, it is not exposed through the system API, and there is
        no obvious way to shut it down.

        '''
        super(RiakStorage, self).close()
        self.connection = None



'''
Tools for storage implementations.

Your use of this software is governed by your license agreement.

Copyright 2012-2015 Diffeo, Inc.
'''

import uuid
import itertools
from kvlayer._exceptions import StorageClosed, BadKey, SerializationError


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)


def _requires_connection(func):
    '''
    Decorator for methods on any implementation of AbstractStorage.
    Raises StorageClosed when self._connected is not True.
    '''
    def wrapped_func(self, *args, **kwargs):
        if not self._connected:
            raise StorageClosed()
        return func(self, *args, **kwargs)
    return wrapped_func


def split_uuids(uuid_str):
    '''
    DEPRECATED. Only used in Cassandra backend.
    '''
    if '\0' in uuid_str:
        return uuid_str.split('\0')
    return map(lambda s: uuid.UUID(hex=''.join(s)), grouper(uuid_str, 32))


def join_uuids(*uuids):
    '''
    DEPRECATED. Only used in Cassandra backend.
    '''
    if not uuids or uuids[0] == '':
        uuid_str = b''
    else:
        uuid_str = ''.join(map(lambda x: x.hex, uuids))
    return uuid_str


class batches(object):
    '''
    return lists of length n drawn from an iterable.  The last batch
    may be shorter than n.
    '''
    def __init__(self, iterable, n):
        self.items = iter(iterable)
        self.n = n

    def __iter__(self):
        batch = []
        for i in self.items:
            batch.append(i)
            if len(batch) == self.n:
                yield batch
                batch = []
        if batch:
            yield batch



"""Configuration parameters for kvlayer.

These collectively implement the `yakonfig.Configurable` interface.

-----

This software is released under an MIT/X11 open source license.

Copyright 2012-2014 Diffeo, Inc.

"""
from __future__ import absolute_import

from kvlayer._client import STORAGE_CLIENTS
from yakonfig import ConfigurationError

config_name = 'kvlayer'
default_config = dict(
    storage_type = 'local',
    connection_pool_size = 2,
    max_consistency_delay = 120,
    replication_factor = 1,
    thrift_framed_transport_size_in_mb = 15,
    encoder = 'ascii_percent',
    )

def add_arguments(parser):
    '''Add command line arguments to an argparse.ArgumentParser instance.'''
    parser.add_argument('--app-name',
                        help='name of app for namespace prefixing')
    parser.add_argument('--namespace',
                        help='namespace for prefixing table names')
    parser.add_argument('--storage-type',
                        help='backend type for kvlayer, e.g. "local" or "accumulo"')
    parser.add_argument('--storage-address', action='append',
                        dest='storage_addresses', metavar='HOST:PORT',
                        help='network addresses for kvlayer, can be repeated')
    parser.add_argument('--username', help='username for kvlayer accumulo')
    parser.add_argument('--password', help='password for kvlayer accumulo')

runtime_keys = dict(
    app_name = 'app_name',
    namespace = 'namespace',
    username = 'username',
    password = 'password',
    storage_type = 'storage_type',
    storage_addresses = 'storage_addresses',
    connection_pool_size = 'connection_pool_size',
    max_consistency_delay = 'max_consistency_delay',
    replication_factor = 'replication_factor',
    thrift_framed_transport_size_in_mb = 'thrift_framed_transport_size_in_mb',
    log_stats = 'log_stats',
    log_stats_interval_ops = 'log_stats_interval_ops',
    log_stats_interval_seconds = 'log_stats_interval_seconds',

    # these support testing and aren't exposed as command-line arguments
    kvlayer_filename = 'filename',
    kvlayer_copy_to_filename = 'copy_to_filename',
)

def discover_config(config, prefix):
    '''Look out in the world to find kvlayer configuration.

    This will generally try to populate the `storage_type` and
    `storage_addresses` parameters, if a reasonable default for those
    can be found.

    '''

    # If the caller specified a top-level storage_addresses, stop
    # (Even if they didn't specify storage_type)
    if 'storage_addresses' in config:
        return

    # Or, if they specified both storage_type and a subconfig with
    # storage_addresses, stop
    storage_type = config.get('storage_type', None)
    if (storage_type is not None and
        storage_type in config and
        'storage_addresses' in config[storage_type]):
        return

    # Otherwise we can try to do discovery (either specifically on
    # storage_type, or on the alphabetically-first thing that has
    # anything)

    # Do we have a storage_addresses for some storage_type already?
    for name in sorted(STORAGE_CLIENTS.iterkeys()):
        if storage_type is not None and storage_type != name:
            continue
        if name not in config:
            continue
        if 'storage_addresses' not in config[name]:
            continue
        # We found something!
        if storage_type is None:
            config['storage_type'] = name
        return

    # Is there some storage_type that can find itself?
    for name in sorted(STORAGE_CLIENTS.iterkeys()):
        if storage_type is not None and storage_type != name:
            continue
        impl = STORAGE_CLIENTS[name]
        if not hasattr(impl, 'discover_config'):
            continue
        sub_config = config.get(name, {})
        impl.discover_config(sub_config, prefix + '.' + name)
        if 'storage_addresses' not in sub_config:
            continue
        # Again, we found something!
        config[name] = sub_config
        if storage_type is None:
            config['storage_type'] = name
        return


def check_config(config, name):
    if 'storage_type' not in config:
        raise ConfigurationError('{} must have a storage_type'
                                 .format(name))
    if config['storage_type'] not in STORAGE_CLIENTS:
        raise ConfigurationError('invalid {} storage_type {}'
                                 .format(name, config['storage_type']))
    if 'namespace' not in config:
        raise ConfigurationError('{} requires a namespace'.format(name))
    if 'app_name' not in config:
        raise ConfigurationError('{} requires an app_name'.format(name))



#!/usr/bin/env python
'''Performance tests for kvlayer.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

This contains various performance-oriented tests that print statistics
to standard output.  If run as a standalone program, it takes no
configuration options, but it does look for ``config_*.yaml`` files in
the test directory and runs tests against every backend that it can
find configuration for.  Returns 1 if any backend encountered an
exception, or 0 otherwise.

'''
from __future__ import absolute_import, division, print_function
import argparse
import cProfile
import logging
import multiprocessing
import os
import pstats
import Queue
from cStringIO import StringIO
import sys
import threading
import time
import traceback
import uuid

logger = logging.getLogger(__name__)


try:
    import dblogger
except ImportError:
    dblogger = None
import yakonfig

import kvlayer
from kvlayer._client import STORAGE_CLIENTS


# Straightforward performance-oriented tests

# chop off the end of 15MB to leave room for thrift message overhead
fifteen_MB_minus_overhead = 15 * 2 ** 20 - 256

# TODO: is this actually a well formed test?
def perftest_large_writes(client, item_size=fifteen_MB_minus_overhead,
                          num_items_per_batch=1, num_batches=10):
    client.setup_namespace(dict(t1=2, t2=3))

    u1, u2 = uuid.uuid1(), uuid.uuid1()

    long_string = b' ' * item_size

    ## send ten batches, so we get some averaging
    
    num = num_batches * num_items_per_batch
    total_bytes = num * len(long_string)
    print('Testing large writes ({} rows of {} KB)'
          .format(num, len(long_string)/1024))

    t1 = time.time()

    for rows in xrange(num):
        client.put('t1', ((u1, u2), long_string))
    t2 = time.time()

    dt = t2 - t1

    single_put_bps = total_bytes / dt

    print ('{} single  rows put={:.1f} sec ({:.1f} per sec, {:.1f} KB/s)'
           .format(num, dt, num / dt, total_bytes / (1024 * dt)))

    for _ in xrange(num_batches):
        client.put('t1', *(((u1, u2), long_string) for row in xrange(num_items_per_batch)))
    t3 = time.time()

    dt = t3 - t2
    print ('{} batched rows put={:.1f} sec ({:.1f} per sec, {:.1f} KB/s)'
           .format(num, dt, num / dt, total_bytes / (1024 * dt)))

    batch_put_bps = total_bytes / dt
    return single_put_bps, batch_put_bps


# TODO: is this actually a useful test? it's not parameterized in any
# way (unless there are external setup changes). I think what it's
# effectively testing is how fast we can read and write keys with
# negligible value-data payload.
def perftest_storage_speed(client):
    "return (read bytes per second, wryte bytes per second)"
    client.setup_namespace(dict(t1=2, t2=3))
    num_rows = 10 ** 4
    key_size = 33 # two 16-byte uuid plus splitter, approx 33 bytes
    print('Testing key-only I/O ({} rows) * (~33 byte keys)'.format(num_rows))
    t1 = time.time()
    client.put('t1', *[((uuid.uuid4(), uuid.uuid4()), b'')
                       for i in xrange(num_rows)])
    t2 = time.time()
    results = list(client.scan('t1', batch_size=num_rows))
    t3 = time.time()
    put_rate = float(num_rows) / (t2 - t1)
    get_rate = float(num_rows) / (t3 - t2)
    print('{} rows '
          'put={:.1f} sec ({:.1f} per sec) '
          'get={:.1f} sec ({:.1f} per sec)'
          .format(num_rows, (t2 - t1), put_rate, (t3 - t2), get_rate))
    assert num_rows == len(results), ('wanted %s rows, got %s', num_rows, len(results))

    num_bytes = num_rows * key_size
    return num_bytes / (t3 - t2), num_bytes / (t2 - t1)


# The giant complex multiprocessing-based concurrent performance test

def run_many(operation, task_generator, timeout, num_workers=1,
             class_config=None, profile=False):
    '''Run some operation function in parallel under multiprocessing.

    Runs `num_workers` child processes and sends them tasks from
    `task_generator`.  This is a generator function that yields the
    results.

        for response from run_many(do_search, queries, num_workers=10,
                                   timeout=5):
            validate(response)

    `operation` is a callable of two parameters, a task from
    `task_generator` and the output queue `o_queue`.  Alternately,
    `operation` can be a class object (or other callable that returns
    a callable of this form), and `class_config` will be passed to it.

    :param operation: callable to do work in the worker
    :param task_generator: iterable of input parameters to `operation`
    :param int timeout: maximum run time, in seconds
    :param int num_workers: number of child processes to run in parallel
    :param class_config: if not :const:`None`, expect `operation` to
      be a class, and pass this to its constructor
    :param bool profile: if true, run workers under a profiler
      and write to pid-specific files in the current directory
    :return: 0 on success, 1 on child process exception, 2 on timeout

    '''

    if num_workers == 1:
        for resp in run_local(operation, task_generator, timeout, num_workers, class_config, profile):
            yield resp
        return

    task_generator = iter(task_generator)

    manager = multiprocessing.Manager()
    tasks_remaining = manager.Event()
    tasks_remaining.set()

    # make I/O queues with no size constraint
    i_queue = manager.Queue()  # pylint: disable=E1101
    o_queue = manager.Queue()  # pylint: disable=E1101

    pool = multiprocessing.Pool(num_workers, maxtasksperchild=1)

    # load up the pool
    async_results = []
    for x in range(num_workers):
        async_res = pool.apply_async(
            worker,
            args=(operation, i_queue, o_queue, tasks_remaining, class_config,
                  profile))
        async_results.append(async_res)

    start_time = time.time()
    while async_results:
        elapsed = time.time() - start_time
        if elapsed > timeout:
            break
        for async_res in async_results:
            try:
                async_res.get(0)
            except multiprocessing.TimeoutError:
                pass
            except Exception:
                logger.error('multiprocessing worker failed', exc_info=True)
                async_results.remove(async_res)
            else:
                assert async_res.ready()
                async_results.remove(async_res)
        while elapsed < timeout:
            elapsed = time.time() - start_time
            try:
                resp = o_queue.get(block=False)
                yield resp
            except Queue.Empty:
                time.sleep(0.2)
                break
        tasks_added = 0
        while (o_queue.qsize() < num_workers * 2 and elapsed < timeout
               and tasks_added < num_workers * 2):
            elapsed = time.time() - start_time
            task = next(task_generator, None)
            if task is None:
                tasks_remaining.clear()
                break
            else:
                i_queue.put(task)
                tasks_added += 1

    pool.close()
    if async_results:
        # Tests timed out, things are still running
        pool.terminate()
    pool.join()


def run_local(operation, task_generator, timeout, num_workers=1,
             class_config=None, profile=False):
    '''Like run_many but locally without multiprocessing'''
    assert num_workers == 1
    #task_generator = iter(task_generator)
    i_queue = Queue.Queue(maxsize=10)
    o_queue = Queue.Queue(maxsize=10)

    tasks_remaining = threading.Event()
    tasks_remaining.set()

    def pusher():
        for task in task_generator:
            i_queue.put(task)
        tasks_remaining.clear()

    push_thread = threading.Thread(target=pusher)
    push_thread.start()

    worker_thread = threading.Thread(
        target=worker,
        args=(operation, i_queue, o_queue, tasks_remaining, class_config,
              profile))

    #start_time = time.time()
    worker_thread.start()

    # TODO: worker signalling so that we can get finer end time than 1.0 second
    while worker_thread.is_alive():
        try:
            resp = o_queue.get(True, 1.0)
            yield resp
        except Queue.Empty:
            pass

    #elapsed = time.time() - start_time


def worker(operation, i_queue, o_queue, tasks_remaining,
           class_config=None, profile=False):
    '''Simple task runner.

    This is intended to run as a single thread under :mod:`multiprocessing`.
    `operation` is a callable of two parameters, a task pulled from
    `i_queue` and the output queue `o_queue`.  Alternately, `operation`
    can be a class object (or other callable that returns a callable of
    this form), and `class_config` will be passed to it.

    :param operation: callable to do work in the worker
    :param i_queue: queue of input tasks
    :type i_queue: :class:`multiprocessing.Queue`
    :param o_queue: queue of results
    :type o_queue: :class:`multiprocessing.Queue`
    :param tasks_remaining: if not set, wait for more work, even on an
      empty queue
    :type tasks_remaining: :class:`multiprocessing.Event`
    :param class_config: if not :const:`None`, expect `operation` to
      be a class, and pass this to its constructor
    :param bool profile: if true, run this worker under a profiler
      and write to a pid-specific file in the current directory

    '''
    logger.info('worker entry')
    pr = None
    if profile:
        logger.info('enabling profiling')
        pr = cProfile.Profile()
        pr.enable()
    if class_config is not None:
        # operation is a class with __call__ so initialize it
        operation = operation(class_config)
    logger.info('worker queue')
    while 1:
        try:
            task = i_queue.get(timeout=0.3)
            operation(task, o_queue)
        except Queue.Empty:
            if tasks_remaining.is_set():
                continue
            else:
                break
        except:
            logger.error('worker error', exc_info=True)
            raise
    if pr:
        pr.disable()
        profstr = pstatstr(pr)
        with open('profile.{}.txt'.format(os.getpid()), 'a') as pf:
            pf.write(profstr)


def pstatstr(prof):
    profout = StringIO()
    profout.write('# {} {!r}\n\n'.format(time.strftime('%Y%m%d_%H%M%S'), sys.argv))
    ps = pstats.Stats(prof, stream=profout)
    ps.sort_stats('cumulative', 'calls')
    ps.print_stats()
    profout.write('\n\tfunction callers\n')
    ps.print_callers()
    profout.write('\n\tfunction callees\n')
    ps.print_callees()
    return profout.getvalue()


class simple_app(object):
    def __init__(self, config):
        logger.info('thing init')
        self.client = config.get('client',None) or kvlayer.client()
        self.client.setup_namespace(dict(t1=2))
        self.item_size = config.pop('item_size', fifteen_MB_minus_overhead)
        self.long_string = b' ' * self.item_size
        self.num_batches = config.pop('num_batches', 10)
        self.num_items_per_batch = config.pop('num_items_per_batch', 1)
        self.num_items = self.num_batches * self.num_items_per_batch


class random_inserts(simple_app):
    '''Worker function that inserts 1 MB keys in a table.'''
    def __call__(self, u, o_queue):
        for _ in xrange(self.num_batches):
            keys = [(u, uuid.uuid4())
                    for _ in xrange(self.num_items_per_batch)]
            try:
                self.client.put('t1', *((key, self.long_string) for key in keys))
                for key in keys:
                    o_queue.put(key)
            except Exception, exc:
                logger.critical('failed', exc_info=True)
                sys.exit(str(exc))


class many_gets(simple_app):
    '''Worker function that retrieves keys from a table.'''
    def __call__(self, key, o_queue):
        try:
            assert list(self.client.get('t1', key))
            o_queue.put((True, key))
        except Exception, exc:
            logger.critical('failed', exc_info=True)
            sys.exit(str(exc))


def perftest_throughput_insert_random(num_workers=4,
                                      profile=False, 
                                      item_size=fifteen_MB_minus_overhead, 
                                      num_items_per_batch=1, 
                                      num_batches=10,
                                      client=None,
                                      ):
    '''Measure concurrent write throughput writing data to a table.'''
    if client is None:
        client = kvlayer.client()
    if client._config.get('storage_type') == 'accumulo':
        import struct
        client.setup_namespace(dict(t1=2))
        step = ((0x7fffffff // 20) * 2) + 1
        splits = [struct.pack('>I', i) for i in xrange(step, 0x0ffffffff, step)]
        logger.info('accumulo splits=%r', splits)
        client.conn.client.addSplits(client.conn.login, client._ns('t1'), splits)

    num_inserts = num_items_per_batch * num_batches
    total_inserts = num_workers * num_inserts
    task_generator = (uuid.uuid4() for x in xrange(num_workers))
    class_config = dict(
        item_size=item_size,
        num_items_per_batch=num_items_per_batch, 
        num_batches=num_batches,
        )
    if num_workers == 1:
        class_config['client'] = client
    start_time = time.time()
    ret_vals = list(run_many(random_inserts, task_generator,
                             timeout=total_inserts * 5,
                             class_config=class_config,
                             num_workers=num_workers,
                             profile=profile))


    elapsed = time.time() - start_time
    assert len(ret_vals) == total_inserts, (len(ret_vals), num_workers, num_batches, num_items_per_batch)
    total_bytes = item_size * total_inserts
    rate = total_inserts / elapsed
    print(
        'parallel {} workers, {} batches, {} items per batch, '
        '{} bytes per item, '
        '{} inserts ({:.4f} MB) written in {:.1f} seconds --> '
        '{:.1f} items/sec, {:.4f} MB/s'
        .format(
            num_workers, num_batches, num_items_per_batch, item_size,
            total_inserts, total_bytes / 2**20, elapsed, 
            rate, total_bytes / (2**20 * elapsed)))
    sys.stdout.flush()
    return ret_vals, (total_bytes / elapsed)


def perftest_throughput_many_gets(ret_vals=[], num_workers=4,
                                  item_size=fifteen_MB_minus_overhead, 
                                  num_items_per_batch=1, 
                                  num_batches=10,
                                  profile=False,
                                  client=None):
    '''Measure concurrent read throughput reading data from a table.'''
    class_config = dict(
        item_size=item_size,
        num_items_per_batch=num_items_per_batch, 
        num_batches=num_batches,
        )
    if (num_workers == 1) and (client is not None):
        class_config['client'] = client
    start_time = time.time()
    count = 0
    for (found, u) in run_many(many_gets, ret_vals,
                               class_config=class_config,
                               num_workers=num_workers,
                               timeout=len(ret_vals) * 5,
                               profile=profile):
        assert found
        count += 1
    elapsed = time.time() - start_time
    assert count == len(ret_vals)

    num_inserts = num_items_per_batch * num_batches
    total_inserts = num_workers * num_inserts
    total_bytes = item_size * total_inserts
    rate = total_inserts / elapsed

    rate = count / elapsed
    print(
        'parallel {} workers, {} batches, {} items per batch, '
        '{} bytes per item, '
        '{} items ({:.4f} MB) read in {:.1f} seconds --> '
        '{:.1f} items/sec, {:.4f} MB/s'
        .format(
            num_workers, num_batches, num_items_per_batch, item_size,
            count, total_bytes/2**20, elapsed, rate,
            total_bytes / (2**20 * elapsed)
        ))
    sys.stdout.flush()
    return total_bytes / elapsed


def run_perftests(num_workers=4, 
                  item_size=fifteen_MB_minus_overhead,
                  num_items_per_batch=1,
                  num_batches=10,
                  profile=False,
                  splitter='\t',
                  out=None):
    '''Run all of the performance tests.

    Must be run in a :mod:`yakonfig` context.

    :return: 0 on success, 1 on any exception

    '''
    if out is None:
        out = sys.stdout
    rc = 0
    name = yakonfig.get_global_config('kvlayer')['storage_type']
    print('Running tests on backend "{}"'.format(name))
    header = ['#num_workers', 'item_size', 'num_items_per_batch', 'num_batches']
    vals = [num_workers, item_size, num_items_per_batch, num_batches]
    try:
        client = kvlayer.client()
        client.delete_namespace()
        single_put_bps, batch_put_bps = perftest_large_writes(
            client, item_size=item_size, 
            num_items_per_batch=num_items_per_batch,
            num_batches=num_batches)
        header.extend(['single_put_bps', 'batch_put_bps'])
        vals.extend([single_put_bps, batch_put_bps])
        client.delete_namespace()
        key_read_bps, key_write_bps = perftest_storage_speed(client)
        header.extend(['key_read_bps', 'key_write_bps'])
        vals.extend([key_read_bps, key_write_bps])
        client.delete_namespace()
        if name not in ['filestorage']:
            ret_vals, insert_bps = perftest_throughput_insert_random(
                num_workers=num_workers,
                item_size=item_size,
                num_items_per_batch=num_items_per_batch,
                num_batches=num_batches,
                profile=profile,
                client=client,
            )
            get_bps = perftest_throughput_many_gets(
                ret_vals=ret_vals,
                num_workers=num_workers,
                item_size=item_size,
                num_items_per_batch=num_items_per_batch,
                num_batches=num_batches,
                profile=profile,
                client=client,
            )
            header.extend(['insert_bps', 'get_bps'])
            vals.extend([insert_bps, get_bps])
        client.delete_namespace()
        client.close()
    except Exception:
        traceback.print_exc()
        rc = 1
    print('')
    out.write(splitter.join(header))
    out.write('\n')
    out.write(splitter.join(map(str, vals)))
    out.write('\n')
    out.flush()
    return rc


def run_all_perftests(redis_address=None, clientlist=None):
    '''Run all of the performance tests, on every backend.

    This is intended to be a fully-automated answer for
    :program:`kvlayer_test`.

    '''
    rc = 0
    for name in STORAGE_CLIENTS.iterkeys():
        if clientlist is not None:
            if name not in clientlist:
                logger.info('skipping backend %r', name)
                continue
        elif name in ['cassandra', 'accumulo', 'postgres', 'cborproxy']:
            logger.info('skipping backend %r', name)
            continue
        config = os.path.join(os.path.dirname(__file__),
                              'config_{}.yaml'.format(name))
        if not os.path.exists(config):
            continue
        params = {'app_name': 'kvlayer_performance',
                  'namespace': 'ns' + uuid.uuid1().hex}
        if name == 'filestorage':
            params['kvlayer_filename'] = os.tmpnam()
        if name == 'redis' and redis_address is not None:
            params['storage_addresses'] = [redis_address]
        with yakonfig.defaulted_config(
                [kvlayer], filename=config, params=params):
            ret = run_perftests()
            if rc == 0:
                rc = ret
        if name == 'filestorage':
            os.unlink(params['kvlayer_filename'])
    return rc


def main():
    parser = argparse.ArgumentParser(
        description='Run kvlayer performance tests on a single backend.',
        conflict_handler='resolve')
    parser.add_argument('--num-workers', action='append', default=[], type=int)
    parser.add_argument('--item-size', action='append', default=[], type=int, 
                        help='size of the items to push in the large writes test, '
                        'defaults to maximum size per record in thrift RPC server '
                        'example, i.e. 15MB minus a bit of overhead')
    parser.add_argument('--num-items-per-batch', action='append', default=[], type=int, 
                        help='number of items per batch in the large writes test, '
                        'defaults to 1')
    parser.add_argument('--num-batches', default=10, type=int, 
                        help='number of batches in the large writes test, '
                        'defaults to 10')
    parser.add_argument('--profile', action='store_true')
    parser.add_argument('--shutdown-proxies', action='store_true')
    parser.add_argument('--out', default=None, help='file to append results to')
    modules = [yakonfig]
    if dblogger:
        modules.append(dblogger)
    modules.append(kvlayer)
    args = yakonfig.parse_args(parser, modules)

    if args.out:
        out = open(args.out, 'a')
    else:
        out = sys.stdout

    if not args.item_size:
        args.item_size = [fifteen_MB_minus_overhead]
    if not args.num_workers:
        args.num_workers = [1]
    if not args.num_items_per_batch:
        args.num_items_per_batch = [1]

    # return code for sys.exit()
    rc = 0
    for num_workers in args.num_workers:
        for num_items_per_batch in args.num_items_per_batch:
            for item_size in args.item_size:
                rc = run_perftests(
                    num_workers=num_workers,
                    item_size=item_size,
                    num_items_per_batch=num_items_per_batch,
                    num_batches=args.num_batches,
                    profile=args.profile,
                    out=out)

    if args.shutdown_proxies:
        # special feature of CBOR RPC proxy, really for testing only!
        client = kvlayer.client()
        client.shutdown_proxies()
    return rc


if __name__ == '__main__':
    sys.exit(main())



'''
Implementation of AbstractStorage using Postgres

Requires that you have the psycopg2 module installed in your environment:

  easy_install psycopg2
OR
  pip install psycopg2

This software is released under an MIT/X11 open source license.

Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import
import contextlib
import logging
import re

import psycopg2
import psycopg2.pool

from kvlayer._abstract_storage import StringKeyedStorage
from kvlayer._exceptions import ProgrammerError


logger = logging.getLogger(__name__)


# SQL strings in this module use python3 style string.format() formatting
# to substitute the table name into the command.


# this is not precisely right.
# http://www.postgresql.org/docs/9.3/static/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
# super-ascii 'letter' chars are also allowed.
# Perhaps in the all-unicode-all-the-time Python3 re package there will be
# better support for characters classes needed to specify this right.
_psql_identifier_re = re.compile(r'[a-z_][a-z0-9_$]*', re.IGNORECASE)


def _valid_namespace(x):
    return bool(_psql_identifier_re.match(x))


# kv_{namespace}
# table, key, value
_CREATE_TABLE = '''CREATE TABLE kv_{namespace} (
  t text,
  k bytea,
  v bytea,
  PRIMARY KEY (t, k)
);

CREATE FUNCTION upsert_{namespace}(tname TEXT, key BYTEA, data BYTEA)
  RETURNS VOID AS
$$
BEGIN
    LOOP
        -- first try to update the key
        UPDATE kv_{namespace} SET v = data WHERE t = tname AND k = key;
        IF found THEN
            RETURN;
        END IF;
        -- not there, so try to insert the key
        -- if someone else inserts the same key concurrently,
        -- we could get a unique-key failure
        BEGIN
            INSERT INTO kv_{namespace}(t,k,v) VALUES (tname, key, data);
            RETURN;
        EXCEPTION WHEN unique_violation THEN
            -- Do nothing, and loop to try the UPDATE again.
        END;
    END LOOP;
END;
$$
LANGUAGE plpgsql;
'''

_DROP_TABLE = "DROP FUNCTION upsert_{namespace}(TEXT,BYTEA,BYTEA)"
_DROP_TABLE_b = '''DROP TABLE kv_{namespace}'''

_CLEAR_TABLE = '''DELETE FROM kv_{namespace} WHERE t = %s'''

# use cursor.callproc() instead of SELECT query.
# _PUT = '''SELECT upsert_{namespace} (%s, %s, %s);'''

_GET_KV = 'SELECT k, v FROM kv_{namespace} WHERE t=%s'
_GET_K = 'SELECT k FROM kv_{namespace} WHERE t=%s'

_GET_EXACT = ' AND k=%s'
_GET_MIN = ' AND k>=%s'
_GET_MAX = ' AND k<%s'
_SCAN_ORDER = ' ORDER BY k ASC'
_INNER_LIMIT = ' LIMIT %s'

_GET = _GET_KV + _GET_EXACT

_DELETE = '''DELETE FROM kv_{namespace} WHERE t = %s AND k = %s;'''

# unused, we always _DELETE single records by key
# _DELETE_RANGE = ('DELETE FROM kv_{namespace} WHERE t = %s AND k >= %s '
#                  'AND K <= %s;''')

# TODO: use this query to list available namespaces
# select tablename from pg_catalog.pg_tables where tablename like 'kv_%';


MAX_BLOB_BYTES = 15000000


def _cursor_check_namespace_table(cursor, namespace):
    cursor.execute('SELECT 1 FROM pg_tables WHERE tablename ILIKE %s',
                   ('kv_' + namespace,))
    return cursor.rowcount > 0


class PGStorage(StringKeyedStorage):
    def __init__(self, *args, **kwargs):
        '''Initialize a storage instance for namespace.
        uses the single string specifier for a connectionn to a postgres db
http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS
        '''
        super(PGStorage, self).__init__(*args, **kwargs)
        if not _valid_namespace(self._namespace):
            raise ProgrammerError('namespace must match re: %r' %
                                  (_psql_identifier_re.pattern,))
        self.storage_addresses = self._config['storage_addresses']
        if not self.storage_addresses:
            raise ProgrammerError(
                'postgres kvlayer needs config["storage_addresses"]')
        self.connection_pool = psycopg2.pool.SimpleConnectionPool(
            self._config.get('min_connections', 2),
            self._config.get('max_connections', 16),
            self.storage_addresses[0]
        )
        self._scan_inner_limit = int(self._config.get('scan_inner_limit',
                                                      1000))

    @contextlib.contextmanager
    def _conn(self):
        '''Produce a PostgreSQL connection from the pool.

        This also runs a single transaction on that connection.  On
        successful completion, the transaction is committed; if any
        exception is thrown, the transaction is aborted.

        On successful completion the connection is returned to the
        pool for reuse.  If any exception is thrown, the connection
        is closed.

        '''
        conn = self.connection_pool.getconn()
        try:
            with conn:
                yield conn
        finally:
            # This has logic to test whether the connection is closed
            # and/or failed and correctly manages returning it to the
            # pool (or not).
            self.connection_pool.putconn(conn)

    def _namespace_table_exists(self):
        with self._conn() as conn:
            with conn.cursor() as cursor:
                return _cursor_check_namespace_table(cursor, self._namespace)

    def setup_namespace(self, table_names, value_types={}):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.

        :param table_names: Each string in table_names becomes the
        name of a table, and the value must be an integer specifying
        the number of UUIDs in the keys

        :type table_names: dict(str = int)
        '''
        super(PGStorage, self).setup_namespace(table_names, value_types)
        with self._conn() as conn:
            with conn.cursor() as cursor:
                if _cursor_check_namespace_table(cursor, self._namespace):
                    # already exists
                    logger.debug('namespace %r already exists, not creating',
                                 self._namespace)
                    return
                cursor.execute(_CREATE_TABLE.format(namespace=self._namespace))

    def delete_namespace(self):
        '''Deletes all data from namespace.'''
        with self._conn() as conn:
            with conn.cursor() as cursor:
                if not _cursor_check_namespace_table(cursor, self._namespace):
                    logger.debug('namespace %r does not exist, not dropping',
                                 self._namespace)
                    return
                try:
                    cursor.execute(
                        _DROP_TABLE.format(namespace=self._namespace))
                    cursor.execute(
                        _DROP_TABLE_b.format(namespace=self._namespace))
                except:
                    logger.warn('error on delete_namespace(%r)',
                                self._namespace, exc_info=True)

    def clear_table(self, table_name):
        'Delete all data from one table'
        with self._conn() as conn:
            with conn.cursor() as cursor:
                cursor.execute(
                    _CLEAR_TABLE.format(namespace=self._namespace),
                    (table_name,)
                )

    def _put(self, table_name, keys_and_values):
        with self._conn() as conn:
            with conn.cursor() as cursor:
                for (k, v) in keys_and_values:
                    cursor.callproc(
                        'upsert_{namespace}'.format(namespace=self._namespace),
                        (table_name, psycopg2.Binary(k), psycopg2.Binary(v)))

    def _unmarshal_k(self, row, key_spec):
        '''Get the key tuple from a response row.'''
        keyraw = row[0]
        if isinstance(keyraw, buffer):
            keyraw = keyraw[:]
        return self._encoder.deserialize(keyraw, key_spec)

    def _unmarshal_kv(self, row, key_spec):
        '''Get the (key,value) pair from a response row.'''
        val = row[1]
        if isinstance(val, buffer):
            if len(val) > MAX_BLOB_BYTES:
                logger.error('key=%r has blob of size %r over limit of %r',
                             row[0], len(val), MAX_BLOB_BYTES)
                return None
            val = val[:]
        key = self._unmarshal_k(row, key_spec)
        return (key, val)

    def _get(self, table_name, keys):
        cmd = _GET.format(namespace=self._namespace)
        with self._conn() as conn:
            for key in keys:
                with conn.cursor(name='get') as cursor:
                    cursor.execute(cmd, (table_name, psycopg2.Binary(key)))
                    found = False
                    for row in cursor:
                        k = row[0]
                        if k is not None:
                            k = k[:]
                        v = row[1]
                        if v is not None:
                            v = v[:]  # un-bufferify
                            found = True
                            yield (k, v)
                    if not found:
                        yield (key, None)

    def _scan(self, table_name, key_ranges):
        for kmin, kmax in (key_ranges or [['', '']]):
            for rkey, rval in self._scan_subscan_kminmax(
                    table_name, kmin, kmax):
                yield rkey, rval

    def _scan_keys(self, table_name, key_ranges):
        for kmin, kmax in (key_ranges or [['', '']]):
            for rkey in self._scan_subscan_kminmax(
                    table_name, kmin, kmax, with_values=False):
                    yield rkey

    def _scan_subscan_kminmax(self, table_name, kmin, kmax, with_values=True):
        prevkey = None
        while True:
            count = 0
            for p in self._scan_kminmax(table_name, kmin, kmax, with_values):
                if with_values:
                    rkey = p[0]
                else:
                    rkey = p
                count += 1
                if rkey != prevkey:
                    # don't double-return an edge value
                    yield p
                prevkey = rkey
            if not self._scan_inner_limit or count < self._scan_inner_limit:
                # we didn't get up to limit, we must be done
                return
            # else, we hit limit, we need to scan for more
            kmin = rkey

    def _scan_kminmax(self, table_name, kmin, kmax, with_values=True):
        if with_values:
            query = _GET_KV
        else:
            query = _GET_K
        query = query.format(namespace=self._namespace)
        args = [table_name]
        if kmin:
            query += _GET_MIN
            args.append(psycopg2.Binary(kmin))
        if kmax:
            query += _GET_MAX
            args.append(psycopg2.Binary(kmax))
        query += _SCAN_ORDER
        if self._scan_inner_limit:
            query += _INNER_LIMIT
            args.append(self._scan_inner_limit)
        with self._conn() as conn:
            with conn.cursor(name='scan') as cursor:
                cursor.execute(query, tuple(args))
                for row in cursor:
                    k = row[0]
                    if k is not None:
                        k = k[:]  # unbufferify
                    if with_values:
                        v = row[1]
                        if v is not None:
                            v = v[:]
                        yield (k, v)
                    else:
                        yield k

    def _delete(self, table_name, keys):
        with self._conn() as conn:
            with conn.cursor() as cursor:
                cursor.executemany(
                    _DELETE.format(namespace=self._namespace),
                    [(table_name, psycopg2.Binary(k)) for k in keys])

    # don't mark this one detatch_on_exception, that would be silly
    def close(self):
        '''
        close connections and end use of this storage client
        '''
        if self.connection_pool:
            try:
                self.connection_pool.closeall()
            finally:
                self.connection_pool = None


# run this to cleanup any cruft from kvlayer unit tests
CLEAN_TESTS = '''
CREATE OR REPLACE FUNCTION clean_tests() RETURNS VOID AS
$$
DECLARE
  argtypes text;
  tat text;
  toid oid;
  pnargs pg_proc%ROWTYPE;
  pt pg_tables%ROWTYPE;
BEGIN
  FOR pnargs IN SELECT * from pg_proc where proname like '%upsert_test_%' LOOP
    SELECT typname FROM pg_type WHERE oid = pnargs.proargtypes[0]
      INTO argtypes;
    FOR i in 1..array_upper(pnargs.proargtypes,1) LOOP
      SELECT typname FROM pg_type WHERE oid = pnargs.proargtypes[i] INTO tat;
      argtypes := argtypes || ',' || tat;
    END LOOP;
    EXECUTE 'DROP FUNCTION ' || pnargs.proname || '(' || argtypes || ');';
  END LOOP;
  FOR pt IN SELECT * FROM pg_tables WHERE tablename LIKE '%kv_test_%' LOOP
    EXECUTE 'DROP TABLE ' || pt.tablename || ';';
  END LOOP;
END
$$ LANGUAGE plpgsql;


SELECT clean_tests();
'''



from __future__ import absolute_import
import uuid
import sys

import pytest

import kvlayer
from kvlayer._local_memory import LocalStorage
import yakonfig

@pytest.yield_fixture
def local_storage():
    config_yaml = """
kvlayer:
    storage_type: local
    storage_addresses: []
    namespace: test
    app_name: test
"""
    with yakonfig.defaulted_config([kvlayer], yaml=config_yaml):
        local_storage = LocalStorage()
        yield local_storage
        local_storage.delete_namespace()

def test_local_storage_singleton(local_storage):
    local_storage.setup_namespace(dict(meta=1))
    keys_and_values = ((uuid.uuid4(),), b'hi')
    local_storage.put('meta', keys_and_values)
    key_range = (keys_and_values[0], keys_and_values[0])
    meta = list(local_storage.scan('meta', key_range))
    assert meta[0][1] == b'hi'
    local_storage2 = LocalStorage()
    local_storage2.setup_namespace(dict(meta=1))
    meta = list(local_storage2.scan('meta', key_range))
    assert meta[0][1] == b'hi'

def test_get(local_storage):
    local_storage.setup_namespace(dict(meta=1))
    value = b'hi2'
    keys = tuple([uuid.uuid4()])
    local_storage.put('meta', tuple((keys, value)))
    res = list(local_storage.get('meta', keys))
    for res_key, res_value in res:
        if res_key == keys:
            assert res_value == value
    assert keys in [rk for rk, rv in res]

def test_get_complex_keys(local_storage):
    local_storage.setup_namespace(dict(meta=2))
    value = b'hi2'
    key = (uuid.uuid4(), uuid.uuid4())
    local_storage.put('meta', tuple((key, value)))
    res = list(local_storage.get('meta', key))
    for res_key, res_value in res:
        if res_key == key:
            assert res_value == value
    assert key in [rk for rk, rv in res]

def test_delete_namespace(local_storage):
    """Test that delete_namespace() actually clears the shared storage"""
    u = (uuid.uuid4(),)

    local_storage.setup_namespace(dict(meta=1))
    local_storage.put('meta', (u, b'hi'))
    assert list(local_storage.get('meta', u)) == [(u, b'hi')]
    local_storage.delete_namespace()

    local_storage2 = LocalStorage()
    local_storage2.setup_namespace(dict(meta=1))
    assert list(local_storage2.get('meta', u)) == [(u, None)]


def test_empty_key(local_storage):
    k = ('', 'a')
    local_storage.setup_namespace({'meta': (str, str)})
    local_storage.put('meta', (k, '1'))
    assert list(local_storage.get('meta', k)) == [(k, '1')]


def test_long_key(local_storage):
    k = ('a', long(5))
    local_storage.setup_namespace({'meta': (str, long)})
    local_storage.put('meta', (k, '1'))
    assert list(local_storage.get('meta', k)) == [(k, '1')]


def test_empty_scan(local_storage):
    k = ('', 'a')
    local_storage.setup_namespace({'meta': (str, str)})
    local_storage.put('meta', (k, '1'))
    assert list(local_storage.scan('meta', (k, k))) == [(k, '1')]


def test_long_scan(local_storage):
    k = ('a', long(5))
    local_storage.setup_namespace({'meta': (str, long)})
    local_storage.put('meta', (k, '1'))

    s, e = ('a', long(1)), ('a', long(10))
    assert list(local_storage.scan('meta', (s, e))) == [(k, '1')]


def test_negative_int_scan(local_storage):
    '''Demontrates improper ordering of signed integers.'''
    k = (-1,)
    local_storage.setup_namespace({'meta': (int,)})
    local_storage.put('meta', (k, '1'))

    s, e = (-2,), (0,)
    assert list(local_storage.scan('meta')) == [(k, '1')]

    assert list(local_storage.scan('meta', (s, e))) == []
    # This should be:
    # assert list(local_storage.scan('meta', (s, e))) == [(k, '1')]



'''
Implementation of AbstractStorage using HBase

Your use of this software is governed by your license agreement.

Copyright 2012-2015 Diffeo, Inc.
'''

import logging
import random

import happybase as hbase

from kvlayer._exceptions import ProgrammerError
from kvlayer._abstract_storage import StringKeyedStorage

logger = logging.getLogger(__name__)


class HBaseStorage(StringKeyedStorage):
    '''
    HBase storage implements kvlayer's AbstractStorage, which
    manages a set of tables as specified in setup_namespace
    '''
    def __init__(self, *args, **kwargs):
        super(HBaseStorage, self).__init__(*args, **kwargs)

        self._connected = False
        addresses = self._config.get('storage_addresses', [])
        if not addresses:
            raise ProgrammerError('config lacks storage_addresses')

        raddr = random.choice(addresses)
        logger.info('connecting to hbase thrift proxy: %r', raddr)
        self._host, self._port = addr_port(raddr, 9090)
        self._conn = None

        self.max_batch_bytes = int(self._config.get('max_batch_bytes',
                                                    10000000))

    config_name = 'hbase'
    default_config = {
        'max_batch_bytes': 10000000,
    }

    @property
    def conn(self):
        # TODO: use hbase.ConnectionPool
        # with self._pool.connection() as connection:
        #   yield connection
        if not self._conn:
            logger.info('connecting to HBase')

            # Real namespaces were supposedly added to HBase 0.96, but
            # HappyBase doesn't seem to support those explicitly. Instead, it
            # namespaces tables with prefixes. One hopes that it uses real
            # namespaces when they're available...
            prefix_sep = '_'
            prefix = '%s%s%s' % (self._app_name, prefix_sep, self._namespace)
            self._conn = hbase.Connection(host=self._host, port=self._port,
                                          table_prefix=prefix,
                                          table_prefix_separator=prefix_sep)
            self._connected = True
        return self._conn

    def _create_table(self, table):
        # See http://goo.gl/X1Tlaf for available options for column families.
        # Explicitly not enabling bloom filters for now because I don't know
        # if they'll make a huge impact. See http://goo.gl/EWprKe for some
        # notes about bloom filters in HBase.
        self.conn.create_table(table, {'d': dict(max_versions=1)})

    def setup_namespace(self, table_names, value_types={}):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace. This operation is idempotent.
        '''
        super(HBaseStorage, self).setup_namespace(table_names, value_types)
        existing_tables = self.conn.tables()
        for table in table_names:
            if table not in existing_tables:
                self._create_table(table)

    def delete_namespace(self):
        '''
        delete all of the tables within namespace
        '''
        for table in self.conn.tables():
            self._delete_table(table)

    def _delete_table(self, table_name):
        # `disable=True` disables the table and then deletes it.
        # (HBase cannot delete enabled tables.)
        self.conn.delete_table(table_name, disable=True)
        logger.debug('deleted table %s_%s_%s', self._app_name, self._namespace, table_name)

    def clear_table(self, table_name):
        self._delete_table(table_name)
        self._create_table(table_name)

    def _put(self, table_name, keys_and_values):
        cur_bytes = 0
        table = self.conn.table(table_name)
        batch = table.batch()
        for key, blob in keys_and_values:
            morelen = len(blob) + len(key)
            if (morelen + cur_bytes) >= self.max_batch_bytes:
                logger.debug('len(blob)=%d + cur_bytes=%d >= '
                             'max_batch_bytes = %d',
                             len(blob), cur_bytes,
                             self.max_batch_bytes)
                logger.debug('pre-emptively sending only what has been '
                             'batched, and will send this item in next batch.')
                batch.send()
                batch = table.batch()
                cur_bytes = 0
            batch.put(key, {'d:d': blob})
            cur_bytes += morelen
        if cur_bytes > 0:
            batch.send()

    def _scan(self, table_name, key_ranges):
        if not key_ranges:
            key_ranges = [['', '']]
        table = self.conn.table(table_name)
        for start_key, stop_key in key_ranges:
            # start_row is inclusive >=
            # end_row is exclusive <
            if start_key or stop_key:
                scanner = table.scan(row_start=start_key, row_stop=stop_key)
            else:
                scanner = table.scan()

            for row in scanner:
                yield (row[0], row[1]['d:d'])

    def _scan_keys(self, table_name, key_ranges):
        if not key_ranges:
            key_ranges = [['', '']]
        table = self.conn.table(table_name)
        for start_key, stop_key in key_ranges:
            # start_row is inclusive >=
            # end_row is exclusive <
            if start_key or stop_key:
                scanner = table.scan(row_start=start_key, row_stop=stop_key,
                                     columns=())
            else:
                scanner = table.scan(columns=())

            for row in scanner:
                yield row[0]

    def _get(self, table_name, keys):
        table = self.conn.table(table_name)
        for key in keys:
            cf = table.row(key, columns=['d:d'])
            val = cf.get('d:d', None)
            yield key, val
            # TODO: is .row or .cells faster?
            # values = table.cells(skey, 'd:d', versions=1)
            # if values:
            #     val = values[0]
            #     yield key, val
            #     num_values += 1
            #     values_size += len(val)
            # else:
            #     yield key, None

    def _delete(self, table_name, keys):
        table = self.conn.table(table_name)
        batch = table.batch()
        for key in keys:
            batch.delete(key)
        batch.send()

    def close(self):
        if self._connected:
            self._connected = False
            self._conn.close()
            self._conn = None

    def __del__(self):
        self.close()


def addr_port(addr, default_port):
    if ':' in addr:
        host, port = addr.split(':')
        return host, int(port)
    else:
        return addr, default_port


def _string_decrement(x):
    if len(x) < 1:
        return None  # None is before all keys, aka negative infinity
    pre = x[:-1]
    post = ord(x[-1])
    if post > 0:
        return pre + chr(post - 1) + '\xff'
    else:
        pre = _string_decrement(pre)
        if pre is None:
            return None
        return pre + '\xff'



'''Key tuple encoder using ASCII text with percent escaping.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

.. autoclass:: AsciiPercentEncoder

'''
from __future__ import absolute_import
import uuid

from kvlayer.encoders.base import Encoder
from kvlayer._exceptions import SerializationError


class AsciiPercentEncoder(Encoder):
    '''Key tuple encoder using ASCII text with percent escaping.

    This serializes :class:`~uuid.UUID`, :class:`int`, and :class:`long`
    to 8-byte ASCII hex strings.  :class:`str` escapes ``%`` to ``%25``
    and null bytes to ``%00``.  Then the tuple of byte strings from
    these encodings is joined with null bytes between them.

    This encoding produces readable keys in most cases, but is
    fairly inefficient.  Sort order is preserved except for strings
    containing null and percent bytes, and negative integers.

    .. versionchanged:: 0.5
       This was the default key encoder before this version.

    '''
    config_name = 'ascii_percent'

    DELIMITER = '\0'

    def serialize(self, key, spec):
        '''Convert a key tuple to a database-native byte string.

        `key` must be a tuple at most as long as `spec`.  `spec` is a
        tuple of types, and each of the items in `key` must be an
        instance of the corresponding type in `spec`.  If this is
        not the case, :exc:`exceptions.TypeError` will be raised.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if spec is None:
            spec = (None,) * len(key)
        if len(key) > len(spec):
            raise TypeError('key too long, wanted {} parts, got {}'
                            .format(len(spec), len(key)))
        return self.DELIMITER.join(self._serialize_one(frag, typ)
                                   for (frag, typ) in zip(key, spec))

    def _serialize_one(self, frag, typ):
        '''Serialize one fragment of a key tuple.

        If `typ` is :const:`None`, then any type of `frag` is
        acceptable.

        :param frag: part of the key
        :param type typ: expected type of `frag`
        :return: serialized byte string
        :raise exceptions.TypeError: if `frag` has the wrong type
        :raise kvlayer._exceptions.SerializationError: if `typ` isn't
          something this can serialize

        '''
        if typ is not None and not isinstance(frag, typ):
            raise TypeError('expected {} in key but got {} ({!r})'
                            .format(typ, type(frag), frag))
        if isinstance(frag, uuid.UUID) or hasattr(frag, 'hex'):
            return frag.hex
        elif isinstance(frag, (int, long)):
            # format an int to the same number of nybbles as a UUID
            return '{0:032x}'.format(frag)
        elif isinstance(frag, basestring):
            return str(frag.replace('%', '%25').replace('\x00', '%00'))
        elif typ is None:
            raise TypeError('could not serialize {} key fragment ({!r})'
                            .format(type(frag), frag))
        else:
            raise SerializationError('could not serialize {} key fragment'
                                     .format(typ))

    def make_start_key(self, key, spec):
        '''Convert a partial key to the start of a scan range.

        This returns a byte string such that all key tuples greater
        than or equal to `key` :meth:`serialize` to greater than or
        equal to the returned byte string.  If `key` is shorter than
        `spec` then it recognizes all longer key tuples with the same
        prefix.  This generally works like :meth:`serialize`.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        dbkey = self.serialize(key, spec)
        if spec and len(key) > 0 and len(key) < len(spec):
            dbkey += self.DELIMITER
        return dbkey

    def make_end_key(self, key, spec):
        '''Convert a partial key to the end of a scan range.

        This returns a byte string such that all key tuples less than
        or equal to `key` :meth:`serialize` to less than or equal to
        the returned byte string.  If `key` is shorter than `spec`
        then it recognizes all longer key tuples with the same prefix.
        This generally works like :meth:`serialize`.

        :param tuple key: key tuple to serialize
        :param tuple spec: key type tuple
        :return: serialized byte string
        :raise exceptions.TypeError: if a key part has the wrong type, or
          `key` is too long
        :raise kvlayer._exceptions.SerializationError: if a key part has
          a type that can't be serialized

        '''
        if key is None:
            return None
        dbkey = self.serialize(key, spec)
        if spec:
            if len(key) == 0:
                dbkey += '\xff'
            elif len(key) < len(spec):
                dbkey += chr(ord(self.DELIMITER) + 1)
            else:
                dbkey += '\x00'
        else:
            dbkey += '\xff'
        return dbkey

    def deserialize(self, dbkey, spec):
        '''Convert a database-native byte string to a key tuple.

        `dbkey` is expected to have the same format as a call to
        :meth:`serialize` for a key as long as `spec`.  The return
        value will always have the same length as `spec`.

        :param bytes dbkey: key byte string from database
        :param tuple spec: key type tuple
        :return: deserialized key tuple
        :raise exceptions.ValueError: if `dbkey` is wrong
        :raise kvlayer._exceptions.SerializationError: if `spec`
          contains types that cannot be deserialized

        '''
        parts = dbkey.split(self.DELIMITER)
        if len(parts) != len(spec):
            raise SerializationError(
                'tried to split key into {0} parts but got {1}, '
                'from {2!r}').format(len(key_spec), len(parts), key_str)
        return tuple(self._deserialize_one(frag, typ)
                     for frag, typ in zip(parts, spec))
        
    def _deserialize_one(self, frag, typ):
        '''Deserialize one fragment of a serialized string.

        :param frag: part of the serialized key
        :param type typ: expected type of the result
        :return: instance of `typ`
        :raise exceptions.ValueError: if `frag` doesn't actually
          deserialize to `typ`
        :raise kvlayer._exceptions.SerializationError: if `typ` isn't
          something this can deserialize
        '''
        if typ is int:
            return int(frag, 16)
        elif typ is long:
            return long(frag, 16)
        elif typ == (int, long):
            return long(frag, 16)
        elif typ is uuid.UUID:
            return uuid.UUID(hex=frag)
        elif typ is str:
            return frag.replace('%00', '\x00').replace('%25', '%')
        else:
            raise SerializationError('could not deserialize {} key fragment'
                                     .format(typ))



#!/usr/bin/env python

import os
import fnmatch
import subprocess

## prepare to run PyTest as a command
from distutils.core import Command

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'kvlayer'
URL = 'http://diffeo.com'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
DESC = 'table-oriented abstraction layer over key-value stores'


def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
        )
    return open(file_path).read()


def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        results.extend(os.path.join(base, f) for f in goodfiles)
    return results


def recursive_glob_with_tree(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        one_dir_results = []
        for f in goodfiles:
            one_dir_results.append(os.path.join(base, f))
        results.append((base, one_dir_results))
    return results


def _myinstall(pkgspec):
    subprocess.check_call(['pip', 'install', pkgspec])


class PyTest(Command):
    '''run py.test'''

    description = 'runs py.test to execute all tests'

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if self.distribution.install_requires:
            for ir in self.distribution.install_requires:
                _myinstall(ir)
        if self.distribution.tests_require:
            for ir in self.distribution.tests_require:
                _myinstall(ir)

        errno = subprocess.call(['py.test', '-n', '3', '-s', 'kvlayer/tests', '--runslow', '--runperf'])
        raise SystemExit(errno)

setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    long_description=read_file('README.md'),
    author=AUTHOR,
    license='MIT/X11 license http://opensource.org/licenses/MIT',
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    cmdclass={
        'test': PyTest,
    },
    # We can select proper classifiers later
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',  ## MIT/X11 license http://opensource.org/licenses/MIT
    ],
    install_requires=[
        'yakonfig >= 0.6.0',
        'pycassa >= 1.10',
        'pyaccumulo >= 1.5.0.5',
        'pyyaml',
        'cql',
        'redis',
    ],
    extras_require={
        'Postgres': ['psycopg2'],
        'Riak': ['riak'],
        'S3': ['boto'],
        'unittest': ['pytest', 'pytest-diffeo'],
    },
    entry_points={
        'console_scripts': [
            'kvlayer = kvlayer._client:main',
            'kvlayer_test = kvlayer.tests.run:main',
            'kvlayer_performance = kvlayer.tests.performance:main',
        ]
    },
    include_package_data=True,
)



'''Implementation of AbstractStorage using Apache Accumulo; requires the
Accumulo Thrift Proxy that is available in Accumulo 1.4.4+

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''

from __future__ import absolute_import
import logging
import random
import re
import struct

from pyaccumulo import Accumulo, Mutation, Range, BatchWriter
from pyaccumulo.iterators import SummingCombiner
from pyaccumulo.proxy.ttypes import IteratorScope, \
    AccumuloSecurityException, IteratorSetting

from kvlayer._abstract_storage import COUNTER, StringKeyedStorage
from kvlayer._decorators import retry
from kvlayer._exceptions import ProgrammerError

logger = logging.getLogger(__name__)


class AStorage(StringKeyedStorage):
    '''Accumulo :mod:`kvlayer` backend.

    Inside Accumulo, all key and value tuples are normalized to
    strings.  Each key is stored as a row value, and column families
    and qualifiers are empty strings.

    If a table has type :class:`~kvlayer._abstract_storage.COUNTER`,
    the Accumulo ``SummingCombiner`` is applied to the table.  Values
    are stored as 8-byte big-endian signed integers and added
    server-side.  This makes :meth:`increment` be faster and atomic,
    since it does not need to get the old values to cause addition to
    happen.  However, it also means that :meth:`put` on these tables
    makes two mutations to delete the old value and then start the new
    sum.

    '''
    def __init__(self, *args, **kwargs):
        super(AStorage, self).__init__(*args, **kwargs)

        self._connected = False
        addresses = self._config.get('storage_addresses', [])
        if not addresses:
            raise ProgrammerError('config lacks storage_addresses')

        def str_to_pair(address):
            if ':' not in address:
                return (address, 50096)
            else:
                h, p = address.split(':')
                return (h, int(p))
        self._addresses = map(str_to_pair, addresses)

        # The following are all parameters to the accumulo
        # batch interfaces.
        self._max_memory = self._config.get('accumulo_max_memory', 1000000)
        self._timeout_ms = self._config.get('accumulo_timeout_ms', 30000)
        self._threads = self._config.get('accumulo_threads', 10)
        self._latency_ms = self._config.get('accumulo_latency_ms', 10)

        # Acumulo storage requires a username and password
        self._user = self._config.get('username', None)
        self._password = self._config.get('password', None)
        if not self._user and self._password:
            raise ProgrammerError('accumulo storage requires '
                                  'username/password')

        self.thrift_framed_transport_size_in_mb = \
            self._config['thrift_framed_transport_size_in_mb']

        self._conn = None

    @property
    def conn(self):
        if not self._conn:
            host, port = random.choice(self._addresses)
            logger.debug('connecting to Accumulo %s:%d',
                         host, port)
            self._conn = Accumulo(host, port,
                                  self._user, self._password)
            self._connected = True
        return self._conn

    def _ns(self, table):
        '''
        accumulo does not have "namespaces" as a concept per se, so we
        achieve the same effect by creating tables with the namespace
        string appended to the end of the table name
        '''
        return '%s_%s_%s' % (self._app_name, self._namespace, table)

    def _create_table(self, table):
        ns_table = self._ns(table)
        if self.conn.table_exists(ns_table):
            logger.debug('table %s exists, not modifying', ns_table)
            return
        logger.debug('creating %s', ns_table)

        self.conn.create_table(ns_table)
        self.conn.client.setTableProperty(self.conn.login,
                                          ns_table,
                                          'table.bloom.enabled',
                                          'true')

        if self._value_types.get(table, str) is COUNTER:
            i = SummingCombiner(encoding_type='FIXEDLEN')
            scopes = set([IteratorScope.SCAN, IteratorScope.MINC,
                          IteratorScope.MAJC])
            i.attach(self.conn, ns_table, scopes)

    def value_to_str(self, value, value_type):
        if value is None:
            return None
        # LongCombiner expects a big-endian 8-byte int
        if value_type is COUNTER:
            return struct.pack('>q', value)
        return super(AStorage, self).value_to_str(value, value_type)

    def str_to_value(self, value, value_type):
        if value is None:
            return None
        if value_type is COUNTER:
            return struct.unpack('>q', value)[0]
        return super(AStorage, self).str_to_value(value, value_type)

    def setup_namespace(self, table_names, value_types={}):
        '''creates tables in the namespace.  Can be run multiple times with
        different table_names in order to expand the set of tables in
        the namespace.
        '''
        super(AStorage, self).setup_namespace(table_names, value_types)
        for table in table_names:
            if not self.conn.table_exists(self._ns(table)):
                self._create_table(table)

    def delete_namespace(self):
        '''
        delete all of the tables within namespace
        '''
        logger.debug('getting list of tables')
        tables = self.conn.list_tables()
        logger.debug('finding tables to delete from %s: %r',
                     self._ns(''), tables)
        tables_to_delete = [x for x in tables if re.search(self._ns(''), x)]
        for table in tables_to_delete:
            self.conn.delete_table(table)

    def clear_table(self, table_name):
        self.conn.delete_table(self._ns(table_name))
        self._create_table(table_name)

    @retry([AccumuloSecurityException])
    def _put(self, table_name, keys_and_values, counter_deletes=True):
        cur_bytes = 0
        max_bytes = self.thrift_framed_transport_size_in_mb * 2 ** 19
        batch_writer = BatchWriter(conn=self.conn,
                                   table=self._ns(table_name),
                                   max_memory=self._max_memory,
                                   latency_ms=self._latency_ms,
                                   timeout_ms=self._timeout_ms,
                                   threads=self._threads)
        try:
            # Because COUNTER is implemented via a summing accumulator,
            # to do a put we need to delete all of the old values before
            # restarting the sum.
            if ((self._value_types.get(table_name, str) is COUNTER and
                 counter_deletes)):
                for key, blob in keys_and_values:
                    mut = Mutation(key)
                    mut.put(cf='', cq='', is_delete=True)
                    batch_writer.add_mutation(mut)
                batch_writer.flush()

            for key, blob in keys_and_values:
                if len(blob) + cur_bytes >= max_bytes:
                    logger.debug('len(blob)=%d + cur_bytes=%d >= '
                                 'thrift_framed_transport_size_in_mb/2 = %d',
                                 len(blob), cur_bytes, max_bytes)
                    logger.debug('pre-emptively sending only what has been '
                                 'batched, and will send this item in next '
                                 'batch.')
                    batch_writer.flush()
                    cur_bytes = 0
                cur_bytes += max_bytes

                mut = Mutation(key)
                mut.put(cf='', cq='', val=blob)
                batch_writer.add_mutation(mut)
        finally:
            batch_writer.close()

    def _scan(self, table_name, key_ranges):
        for kv in self._do_scan(table_name, key_ranges, keys_only=False):
            yield kv

    def _scan_keys(self, table_name, key_ranges):
        for kv in self._do_scan(table_name, key_ranges, keys_only=True):
            yield kv

    def _do_scan(self, table_name, key_ranges, keys_only):
        iterators = []
        if keys_only:
            iterators.append(IteratorSetting(
                name='SortedKeyIterator', priority=10,
                iteratorClass='org.apache.accumulo.core.iterators.'
                'SortedKeyIterator', properties={}))
        if not key_ranges:
            key_ranges = [['', '']]
        for start_key, stop_key in key_ranges:
            specific_key_range = bool(start_key or stop_key)
            if specific_key_range:
                # Accumulo treats None as a negative-infinity or
                # positive-infinity key as needed for starts and ends
                # of ranges.
                #
                # pyaccumulo has a bug at present 20140228_171555
                # which causes it to never do a '>=' scan, so we must
                # decrement the start key to include the start key
                # which necessary for a get() operation.
                # https://github.com/accumulo/pyaccumulo/issues/14
                if start_key:
                    start_key = _string_decrement(start_key)
                key_range = Range(srow=start_key, erow=stop_key,
                                  sinclude=True, einclude=True)
                scanner = self.conn.scan(self._ns(table_name),
                                         scanrange=key_range,
                                         iterators=iterators)
            else:
                scanner = self.conn.scan(self._ns(table_name),
                                         iterators=iterators)

            for row in scanner:
                if keys_only:
                    yield row.row
                else:
                    yield (row.row, row.val)

    def _get(self, table_name, keys):
        for key in keys:
            gen = self._do_scan(table_name, [(key, key)], keys_only=False)
            v = None
            for kk, vv in gen:
                if kk == key:
                    v = vv
            yield key, v

    def close(self):
        self._connected = False
        if hasattr(self, 'pool') and self.pool:
            self.pool.dispose()

    def __del__(self):
        self.close()

    def _delete(self, table_name, keys):
        batch_writer = BatchWriter(conn=self.conn,
                                   table=self._ns(table_name),
                                   max_memory=self._max_memory,
                                   latency_ms=self._latency_ms,
                                   timeout_ms=self._timeout_ms,
                                   threads=self._threads)
        try:
            for key in keys:
                mut = Mutation(key)
                mut.put(cf='', cq='', is_delete=True)
                batch_writer.add_mutation(mut)
        finally:
            batch_writer.close()

    def increment(self, table_name, *keys_and_values):
        '''Add values to a counter-type table.

        .. see:: `kvlayer._abstract_storage.AbstractStorage.increment`

        If `table_name` is a :class:`COUNTER` table, then the actual
        summing is done server-side.  This means that concurrent calls
        to this method are atomic against each other: if clients on
        different system increment the same key in the same table by 2
        and 3, respectively, then later reads will return 5.  Behavior
        with concurrent put and delete should be reasonable as well.

        '''
        if self._value_types.get(table_name, str) is COUNTER:
            # Special case for Accumulo: just write the delta values,
            # and the summing combiner will do the math for us
            self.put(table_name, *keys_and_values, counter_deletes=False)
        else:
            super(AStorage, self).increment(table_name, *keys_and_values)


def _string_decrement(x):
    if len(x) < 1:
        return None  # None is before all keys, aka negative infinity
    pre = x[:-1]
    post = ord(x[-1])
    if post > 0:
        return pre + chr(post - 1) + '\xff'
    else:
        pre = _string_decrement(pre)
        if pre is None:
            return None
        return pre + '\xff'
